from unittest import TestCase

from nose.tools import istest, eq_
import iris.analysis

from cis.data_io.gridded_data import GriddedDataList
from cis.aggregation.gridded_collapsor import GriddedCollapsor
from cis.test.utils_for_testing import *
from cis.aggregation.collapse_kernels import aggregation_kernels
from cis.test.util.mock import *
from cis.data_io.gridded_data import make_from_cube


class TestGriddedCollapse(TestCase):
    def setUp(self):
        self.cube = make_from_cube(make_mock_cube())
        self.kernel = iris.analysis.MEAN

    @istest
    def test_collapsing_coordinate_collapses_coordinate(self):
        cube_out = self.cube.collapsed('x', how=self.kernel)

        result = numpy.array([2, 5, 8, 11, 14])

        # There is a small deviation to the weighting correction applied by Iris when completely collapsing
        assert_arrays_almost_equal(result, cube_out[0].data)
        assert_arrays_almost_equal(self.cube.coords('latitude')[0].points, cube_out.coords('latitude')[0].points)

    @istest
    def test_collapsing_coordinate_takes_start_end_but_ignores_them(self):
        cube_out = self.cube.collapsed('x', how=self.kernel)

        result = numpy.array([2, 5, 8, 11, 14])

        # There is a small deviation to the weighting correction applied by Iris when completely collapsing
        assert_arrays_almost_equal(result, cube_out[0].data)
        assert_arrays_almost_equal(self.cube.coords('latitude')[0].points, cube_out.coords('latitude')[0].points)

    @istest
    def test_can_name_variables_by_standard_name(self):
        cube_out = self.cube.collapsed(['longitude', 'latitude'], how=self.kernel)

        result = numpy.array([8.0])

        assert_arrays_almost_equal(result, cube_out[0].data)

    @istest
    def test_can_name_variables_by_variable_name(self):
        cube_out = self.cube.collapsed(['lon', 'lat'], how=self.kernel)

        result = numpy.array([8.0])

        assert_arrays_almost_equal(result, cube_out[0].data)

    @istest
    def test_collapsing_everything_returns_a_single_value(self):
        cube_out = self.cube.collapsed(['x', 'y'], how=self.kernel)

        result = numpy.array([8.0])

        assert_arrays_almost_equal(result, cube_out[0].data)

    @istest
    def test_collapsing_everything_returns_a_single_value_with_missing_values(self):
        self.cube = make_from_cube(make_5x3_lon_lat_2d_cube_with_missing_data())

        cube_out = self.cube.collapsed(['x', 'y'], how=self.kernel)

        # result = numpy.array(8.1538461538461533)
        result = numpy.array(numpy.mean(self.cube.data))

        assert numpy.allclose(result, cube_out[0].data, rtol=1e-2)

    @istest
    def test_aggregating_using_max_kernel_returns_maximums(self):
        self.kernel = iris.analysis.MAX
        cube_out = self.cube.collapsed(['x', 'y'], how=self.kernel)

        result = numpy.array(15)

        assert_arrays_almost_equal(result, cube_out[0].data)

    @istest
    def test_aggregating_using_min_kernel_returns_minimums(self):
        self.kernel = iris.analysis.MIN
        cube_out = self.cube.collapsed(['x', 'y'], how=self.kernel)

        result = numpy.array(1)

        assert_arrays_almost_equal(result, cube_out[0].data)

    @istest
    def test_aggregating_using_std_dev_kernel_returns_sample_standard_deviation(self):
        self.kernel = iris.analysis.STD_DEV
        cube_out = self.cube.collapsed(['y'], how=self.kernel)

        result = numpy.array([numpy.sqrt(22.5), numpy.sqrt(22.5), numpy.sqrt(22.5)])

        assert_arrays_almost_equal(result, cube_out[0].data)
        assert_arrays_almost_equal(self.cube.coords('longitude')[0].points, cube_out.coords('longitude')[0].points)

    @istest
    def test_aggregation_on_three_dimensional_grid_with_time(self):
        self.cube = make_from_cube(make_mock_cube(time_dim_length=7))

        cube_out = self.cube.collapsed(['t', 'x', 'y'], how=self.kernel)

        result_data = numpy.array(53)
        assert_arrays_almost_equal(result_data, cube_out[0].data)

    def test_aggregation_over_multidimensional_coord(self):
        self.cube = make_from_cube(make_mock_cube(time_dim_length=7, hybrid_pr_len=5))
        cube_out = self.cube.collapsed(['t', 'x', 'y', 'air_pressure'], how=self.kernel)

        result_data = numpy.array(263)
        assert_arrays_almost_equal(cube_out[0].data, result_data)

    def test_partial_aggregation_over_multidimensional_coord(self):
        # JASCIS-126
        self.cube = make_from_cube(make_mock_cube(time_dim_length=7, hybrid_pr_len=5))
        cube_out = self.cube.collapsed(['t'], how=self.kernel)

        result_data = numpy.array([[[16.0, 17.0, 18.0, 19.0, 20.0],
                                    [51.0, 52.0, 53.0, 54.0, 55.0],
                                    [86.0, 87.0, 88.0, 89.0, 90.0]],

                                   [[121.0, 122.0, 123.0, 124.0, 125.0],
                                    [156.0, 157.0, 158.0, 159.0, 160.0],
                                    [191.0, 192.0, 193.0, 194.0, 195.0]],

                                   [[226.0, 227.0, 228.0, 229.0, 230.0],
                                    [261.0, 262.0, 263.0, 264.0, 265.0],
                                    [296.0, 297.0, 298.0, 299.0, 300]],

                                   [[331.0, 332.0, 333.0, 334.0, 335.0],
                                    [366.0, 367.0, 368.0, 369.0, 370.0],
                                    [401.0, 402.0, 403.0, 404.0, 405.0]],

                                   [[436.0, 437.0, 438.0, 439.0, 440.0],
                                    [471.0, 472.0, 473.0, 474.0, 475.0],
                                    [506.0, 507.0, 508.0, 509.0, 510.0]]], dtype=np.float)

        multidim_coord_points = numpy.array([[300000., 1000000., 1700000.],
                                             [2400000., 3100000., 3800000.],
                                             [4500000., 5200000., 5900000.],
                                             [6600000., 7300000., 8000000.],
                                             [8700000., 9400000., 10100000.]], dtype=np.float)

        assert_arrays_almost_equal(cube_out[0].data, result_data)
        assert_arrays_almost_equal(cube_out[0].coord('surface_air_pressure').points, multidim_coord_points)

    def test_partial_aggregation_over_multidimensional_coord_with_multi_kernel(self):
        # JASCIS-126
        from cis.aggregation.collapse_kernels import MultiKernel, StddevKernel, CountKernel
        self.kernel = MultiKernel('moments', [iris.analysis.MEAN, StddevKernel(), CountKernel()])
        self.cube = make_from_cube(make_mock_cube(time_dim_length=7, hybrid_pr_len=5))
        cube_out = self.cube.collapsed(['t'], how=self.kernel)

        result_data = numpy.array([[[16.0, 17.0, 18.0, 19.0, 20.0],
                                    [51.0, 52.0, 53.0, 54.0, 55.0],
                                    [86.0, 87.0, 88.0, 89.0, 90.0]],

                                   [[121.0, 122.0, 123.0, 124.0, 125.0],
                                    [156.0, 157.0, 158.0, 159.0, 160.0],
                                    [191.0, 192.0, 193.0, 194.0, 195.0]],

                                   [[226.0, 227.0, 228.0, 229.0, 230.0],
                                    [261.0, 262.0, 263.0, 264.0, 265.0],
                                    [296.0, 297.0, 298.0, 299.0, 300]],

                                   [[331.0, 332.0, 333.0, 334.0, 335.0],
                                    [366.0, 367.0, 368.0, 369.0, 370.0],
                                    [401.0, 402.0, 403.0, 404.0, 405.0]],

                                   [[436.0, 437.0, 438.0, 439.0, 440.0],
                                    [471.0, 472.0, 473.0, 474.0, 475.0],
                                    [506.0, 507.0, 508.0, 509.0, 510.0]]], dtype=np.float)

        multidim_coord_points = numpy.array([[300000., 1000000., 1700000.],
                                             [2400000., 3100000., 3800000.],
                                             [4500000., 5200000., 5900000.],
                                             [6600000., 7300000., 8000000.],
                                             [8700000., 9400000., 10100000.]], dtype=np.float)

        assert_arrays_almost_equal(cube_out[0].data, result_data)
        assert_arrays_almost_equal(cube_out[1].data, np.ones(result_data.shape)*10.8012345)
        assert_arrays_almost_equal(cube_out[2].data, np.ones(result_data.shape) * 7)
        assert_arrays_almost_equal(cube_out[0].coord('surface_air_pressure').points, multidim_coord_points)

    def test_partial_aggregation_over_more_than_one_dim_on_multidimensional_coord(self):
        self.cube = make_from_cube(make_mock_cube(time_dim_length=7, hybrid_pr_len=5))
        cube_out = self.cube.collapsed(['t', 'x'], how=self.kernel)

        result_data = numpy.array([[51.0, 52.0, 53.0, 54.0, 55.0],
                                   [156.0, 157.0, 158.0, 159.0, 160.0],
                                   [261.0, 262.0, 263.0, 264.0, 265.0],
                                   [366.0, 367.0, 368.0, 369.0, 370.0],
                                   [471.0, 472.0, 473.0, 474.0, 475.0]], dtype=np.float)

        multidim_coord_points = numpy.array([1000000.,  3100000., 5200000., 7300000., 9400000.], dtype=np.float)

        assert_arrays_almost_equal(cube_out[0].data, result_data)
        assert_arrays_almost_equal(cube_out[0].coord('surface_air_pressure').points, multidim_coord_points)

    def test_partial_aggregation_over_more_than_one_multidimensional_coord(self):
        self.cube = make_from_cube(make_mock_cube(time_dim_length=7, hybrid_pr_len=5, geopotential_height=True))
        cube_out = self.cube.collapsed(['t', 'x'], how=self.kernel)

        result_data = numpy.array([[51.0, 52.0, 53.0, 54.0, 55.0],
                                   [156.0, 157.0, 158.0, 159.0, 160.0],
                                   [261.0, 262.0, 263.0, 264.0, 265.0],
                                   [366.0, 367.0, 368.0, 369.0, 370.0],
                                   [471.0, 472.0, 473.0, 474.0, 475.0]], dtype=np.float)

        altitude_points = result_data + 9

        surface_air_pressure_points = numpy.array([1000000., 3100000., 5200000., 7300000., 9400000.], dtype=np.float)

        assert_arrays_almost_equal(cube_out[0].data, result_data)
        assert_arrays_almost_equal(cube_out[0].coord('surface_air_pressure').points, surface_air_pressure_points)
        assert_arrays_almost_equal(cube_out[0].coord('altitude').points, altitude_points)

    def test_partial_aggregation_over_multidimensional_coord_along_middle_of_cube(self):
        # JASCIS-126
        self.cube = make_from_cube(make_mock_cube(time_dim_length=7, hybrid_pr_len=5))
        cube_out = self.cube.collapsed(['x'], how=self.kernel)

        result_data = numpy.array([[[36.0, 37.0, 38.0, 39.0, 40.0],
                                    [41.0, 42.0, 43.0, 44.0, 45.0],
                                    [46.0, 47.0, 48.0, 49.0, 50.0],
                                    [51.0, 52.0, 53.0, 54.0, 55.0],
                                    [56.0, 57.0, 58.0, 59.0, 60.0],
                                    [61.0, 62.0, 63.0, 64.0, 65.0],
                                    [66.0, 67.0, 68.0, 69.0, 70.0]],

                                   [[141.0, 142.0, 143.0, 144.0, 145.0],
                                    [146.0, 147.0, 148.0, 149.0, 150.0],
                                    [151.0, 152.0, 153.0, 154.0, 155.0],
                                    [156.0, 157.0, 158.0, 159.0, 160.0],
                                    [161.0, 162.0, 163.0, 164.0, 165.0],
                                    [166.0, 167.0, 168.0, 169.0, 170.0],
                                    [171.0, 172.0, 173.0, 174.0, 175.0]],

                                   [[246.0, 247.0, 248.0, 249.0, 250.0],
                                    [251.0, 252.0, 253.0, 254.0, 255.0],
                                    [256.0, 257.0, 258.0, 259.0, 260.0],
                                    [261.0, 262.0, 263.0, 264.0, 265.0],
                                    [266.0, 267.0, 268.0, 269.0, 270.0],
                                    [271.0, 272.0, 273.0, 274.0, 275.0],
                                    [276.0, 277.0, 278.0, 279.0, 280.0]],

                                   [[351.0, 352.0, 353.0, 354.0, 355.0],
                                    [356.0, 357.0, 358.0, 359.0, 360.0],
                                    [361.0, 362.0, 363.0, 364.0, 365.0],
                                    [366.0, 367.0, 368.0, 369.0, 370.0],
                                    [371.0, 372.0, 373.0, 374.0, 375.0],
                                    [376.0, 377.0, 378.0, 379.0, 380.0],
                                    [381.0, 382.0, 383.0, 384.0, 385.0]],

                                   [[456.0, 457.0, 458.0, 459.0, 460.0],
                                    [461.0, 462.0, 463.0, 464.0, 465.0],
                                    [466.0, 467.0, 468.0, 469.0, 470.0],
                                    [471.0, 472.0, 473.0, 474.0, 475.0],
                                    [476.0, 477.0, 478.0, 479.0, 480.0],
                                    [481.0, 482.0, 483.0, 484.0, 485.0],
                                    [486.0, 487.0, 488.0, 489.0, 490.0]]], dtype=np.float)

        multidim_coord_points = numpy.array([[700000., 800000., 900000., 1000000., 1100000., 1200000., 1300000.],
                                             [2800000., 2900000., 3000000., 3100000., 3200000., 3300000., 3400000.],
                                             [4900000., 5000000., 5100000., 5200000., 5300000., 5400000., 5500000.],
                                             [7000000., 7100000., 7200000., 7300000., 7400000., 7500000., 7600000.],
                                             [9100000., 9200000., 9300000., 9400000., 9500000., 9600000., 9700000.]],
                                            dtype=np.float)

        assert_arrays_almost_equal(cube_out[0].data, result_data)
        assert_arrays_almost_equal(cube_out[0].coord('surface_air_pressure').points, multidim_coord_points)

    def test_calc_new_dims(self):
        # Cube dims         [0, 1, 2, 3]
        # Coord dims        [0, 1, 2]
        # Dims to collapse      ^
        # Untouched dims    [0, 2]
        # New dims          [0, 1]
        # Local dims        [1]
        # Mirrors: test_partial_aggregation_over_multidimensional_coord_along_middle_of_cube
        eq_(GriddedCollapsor._calc_new_dims([0, 1, 2], {1}), [0, 1])

        # Cube dims         [0, 1, 2, 3]
        # Coord dims        [0, 1, 2]
        # Dims to collapse         ^
        # Untouched dims    [0, 1]
        # New dims          [0, 1]
        # Local dims        [2]
        # Mirrors: test_partial_aggregation_over_multidimensional_coord
        eq_(GriddedCollapsor._calc_new_dims([0, 1, 2], {2}), [0, 1])

        # Cube dims         [0, 1, 2, 3]
        # Coord dims        [0, ., 1, 2]
        # Coord dims        [0, 2, 3]
        # Dims to collapse   ^
        # Untouched dims    [1, 2]
        # New dims          [1, 2]
        # Local dims        [0]
        # Mirrors: test_netCDF_gridded_hybrid_height_partial
        eq_(GriddedCollapsor._calc_new_dims([0, 2, 3], {0}), [1, 2])

        # Test collapsing multiple dimensions
        eq_(GriddedCollapsor._calc_new_dims([0, 1, 2], {0, 1}), [0])

        # Test collapsing multiple dimensions, some of which aren't on the coordinate
        eq_(GriddedCollapsor._calc_new_dims([0, 1, 2], {2, 3}), [0, 1])


class TestGriddedListAggregation(TestCase):
    def setUp(self):
        self.kernel = iris.analysis.MEAN

    def test_collapse_coordinate(self):
        from cis.data_io.gridded_data import GriddedDataList, make_from_cube

        data1 = make_from_cube(make_mock_cube())
        data2 = make_from_cube(make_mock_cube(data_offset=1))
        datalist = GriddedDataList([data1, data2])
        cube_out = datalist.collapsed(['x'], how=self.kernel)

        result1 = numpy.array([2, 5, 8, 11, 14])
        result2 = result1 + 1

        assert isinstance(cube_out, GriddedDataList)

        # There is a small deviation to the weighting correction applied by Iris when completely collapsing
        assert_arrays_almost_equal(result1, cube_out[0].data)
        assert_arrays_almost_equal(result2, cube_out[1].data)
        assert numpy.array_equal(data1.coords('latitude')[0].points, cube_out.coords('latitude')[0].points)

    def test_collapse_vertical_coordinate(self):
        from cis.data_io.gridded_data import GriddedDataList, make_from_cube

        data1 = make_from_cube(make_mock_cube(alt_dim_length=6))
        data2 = make_from_cube(make_mock_cube(alt_dim_length=6, data_offset=1))
        datalist = GriddedDataList([data1, data2])
        cube_out = datalist.collapsed(['z'], how=self.kernel)

        result1 = data1.data.mean(axis=2)
        result2 = result1 + 1

        assert isinstance(cube_out, GriddedDataList)

        # There is a small deviation to the weighting correction applied by Iris when completely collapsing
        assert_arrays_almost_equal(result1, cube_out[0].data)
        assert_arrays_almost_equal(result2, cube_out[1].data)
        assert numpy.array_equal(data1.coords('latitude')[0].points, cube_out.coords('latitude')[0].points)

    def test_collapse_vertical_coordinate_weighted_aggregator(self):
        """
        We use a weighted aggregator, though no weights should be applied since we're only summing over the vertical
        """
        from cis.data_io.gridded_data import GriddedDataList, make_from_cube

        data1 = make_from_cube(make_mock_cube(alt_dim_length=6))
        data2 = make_from_cube(make_mock_cube(alt_dim_length=6, data_offset=1))
        datalist = GriddedDataList([data1, data2])
        cube_out = datalist.collapsed(['z'], how=iris.analysis.SUM)

        result1 = np.sum(data1.data, axis=2)
        result2 = np.sum(data2.data, axis=2)

        assert isinstance(cube_out, GriddedDataList)

        # There is a small deviation to the weighting correction applied by Iris when completely collapsing
        assert_arrays_almost_equal(result1, cube_out[0].data)
        assert_arrays_almost_equal(result2, cube_out[1].data)
        assert numpy.array_equal(data1.coords('latitude')[0].points, cube_out.coords('latitude')[0].points)

    def test_aggregate_mean(self):
        from cis.data_io.gridded_data import GriddedDataList, make_from_cube

        data1 = make_from_cube(make_mock_cube())
        data2 = make_from_cube(make_mock_cube(data_offset=1))
        datalist = GriddedDataList([data1, data2])
        cube_out = datalist.collapsed(['y'], how=self.kernel)

        result1 = numpy.array([7, 8, 9])
        result2 = result1 + 1

        assert isinstance(cube_out, GriddedDataList)

        # There is a small deviation to the weighting correction applied by Iris when completely collapsing
        assert_arrays_almost_equal(result1, cube_out[0].data)
        assert_arrays_almost_equal(result2, cube_out[1].data)

    def test_complete_collapse_one_dim_using_moments_kernel(self):
        self.kernel = aggregation_kernels['moments']
        data1 = make_from_cube(make_5x3_lon_lat_2d_cube_with_missing_data())
        data1.var_name = 'var1'
        data2 = make_from_cube(make_5x3_lon_lat_2d_cube_with_missing_data())
        data2.var_name = 'var2'
        data2.data += 10
        data = GriddedDataList([data1, data2])

        output = data.collapsed(['x'], how=self.kernel)

        expect_mean = numpy.array([[5.5, 8.75, 9]])
        expect_stddev = numpy.array([numpy.sqrt(15), numpy.sqrt(26.25), numpy.sqrt(30)])
        expect_count = numpy.array([[4, 4, 4]])

        assert isinstance(output, GriddedDataList)
        assert len(output) == 6
        mean_1, stddev_1, count_1, mean_2, stddev_2, count_2 = output
        assert mean_1.var_name == 'var1'
        assert stddev_1.var_name == 'var1_std_dev'
        assert count_1.var_name == 'var1_num_points'
        assert mean_2.var_name == 'var2'
        assert stddev_2.var_name == 'var2_std_dev'
        assert count_2.var_name == 'var2_num_points'
        assert_arrays_almost_equal(mean_1.data, expect_mean)
        assert_arrays_almost_equal(mean_2.data, expect_mean + 10)
        assert_arrays_almost_equal(stddev_1.data, expect_stddev)
        assert_arrays_almost_equal(stddev_2.data, expect_stddev)
        assert_arrays_almost_equal(count_1.data, expect_count)
        assert_arrays_almost_equal(count_2.data, expect_count)

    def test_complete_collapse_two_dims_using_moments_kernel(self):
        self.kernel = aggregation_kernels['moments']
        data1 = make_from_cube(make_5x3_lon_lat_2d_cube_with_missing_data())
        data1.var_name = 'var1'
        data2 = make_from_cube(make_5x3_lon_lat_2d_cube_with_missing_data())
        data2.var_name = 'var2'
        data2.data += 10
        data = GriddedDataList([data1, data2])
        output = data.collapsed(['x', 'y'], how=self.kernel)

        expect_mean = numpy.array(7.75)
        expect_stddev = numpy.array(numpy.sqrt(244.25 / 11))
        expect_count = numpy.array(12)

        assert isinstance(output, GriddedDataList)
        assert len(output) == 6
        mean_1, stddev_1, count_1, mean_2, stddev_2, count_2 = output
        assert mean_1.var_name == 'var1'
        assert stddev_1.var_name == 'var1_std_dev'
        assert count_1.var_name == 'var1_num_points'
        assert mean_2.var_name == 'var2'
        assert stddev_2.var_name == 'var2_std_dev'
        assert count_2.var_name == 'var2_num_points'
        # Latitude area weighting means these aren't quite right so increase the rtol.
        assert numpy.allclose(mean_1.data, expect_mean, 1e-3)
        assert numpy.allclose(mean_2.data, expect_mean + 10, 1e-3)
        assert numpy.allclose(stddev_1.data, expect_stddev)
        assert numpy.allclose(stddev_2.data, expect_stddev)
        assert numpy.allclose(count_1.data, expect_count)
        assert numpy.allclose(count_2.data, expect_count)

    def test_partial_aggregation_over_more_than_one_dim_on_multidimensional_coord(self):
        from cis.data_io.gridded_data import GriddedDataList, make_from_cube

        data1 = make_from_cube(make_mock_cube(time_dim_length=7, hybrid_pr_len=5))
        data2 = make_from_cube(make_mock_cube(time_dim_length=7, hybrid_pr_len=5, data_offset=1))
        datalist = GriddedDataList([data1, data2])

        cube_out = datalist.collapsed(['t', 'x'], how=self.kernel)

        result_data = numpy.array([[51.0, 52.0, 53.0, 54.0, 55.0],
                                   [156.0, 157.0, 158.0, 159.0, 160.0],
                                   [261.0, 262.0, 263.0, 264.0, 265.0],
                                   [366.0, 367.0, 368.0, 369.0, 370.0],
                                   [471.0, 472.0, 473.0, 474.0, 475.0]], dtype=np.float)

        multidim_coord_points = numpy.array([1000000., 3100000., 5200000., 7300000., 9400000.], dtype=np.float)

        assert_arrays_almost_equal(cube_out[0].data, result_data)
        assert_arrays_almost_equal(cube_out[1].data, result_data+1)
        assert_arrays_almost_equal(cube_out[0].coord('surface_air_pressure').points, multidim_coord_points)
        assert_arrays_almost_equal(cube_out[1].coord('surface_air_pressure').points, multidim_coord_points)




from netCDF4 import Dataset

from hamcrest import assert_that, greater_than_or_equal_to, less_than_or_equal_to
from nose.tools import raises
import unittest

from cis.cis_main import subset_cmd
from cis.parse import parse_args
from cis.test.integration_test_data import *
from cis.test.integration.base_integration_test import BaseIntegrationTest
from cis.time_util import convert_time_since_to_std_time
from cis.exceptions import CoordinateNotFoundError, NoDataInSubsetError


class TestSubsetIntegration(BaseIntegrationTest):
    def test_GIVEN_single_variable_in_ungridded_file_WHEN_subset_THEN_subsetted_correctly(self):
        variable = valid_aerosol_cci_variable
        filename = valid_aerosol_cci_filename
        lon_min, lon_max = -10, 10
        lat_min, lat_max = 40, 60
        arguments = ['subset', variable + ':' + escape_colons(filename),
                     'x=[%s,%s],y=[%s,%s]' % (lon_min, lon_max, lat_min, lat_max), '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, [variable])

    def test_GIVEN_shape_WHEN_subset_ungridded_data_THEN_subsetted_correctly(self):
        variable = valid_aerosol_cci_variable
        filename = valid_aerosol_cci_filename
        lon_min, lon_max = -10, 10
        lat_min, lat_max = 40, 60
        shape_wkt = "POLYGON((-10 50, 0 60, 10 50, 0 40, -10 50))"
        arguments = ['subset', variable + ':' + escape_colons(filename),
                     'shape=%s' % shape_wkt, '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, [variable])

    def test_GIVEN_single_variable_as_var_name_in_ungridded_file_WHEN_subset_THEN_subsetted_correctly(self):
        variable = valid_aerosol_cci_variable
        filename = valid_aerosol_cci_filename
        lon_min, lon_max = -10, 10
        lat_min, lat_max = 40, 60
        arguments = ['subset', variable + ':' + escape_colons(filename),
                     'lon=[%s,%s],lat=[%s,%s]' % (lon_min, lon_max, lat_min, lat_max), '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, [variable])

    def test_GIVEN_single_variable_in_gridded_file_WHEN_subset_THEN_subsetted_correctly(self):
        variable = valid_hadgem_variable
        filename = valid_hadgem_filename
        lon_min, lon_max = 0, 10
        lat_min, lat_max = 40, 60
        arguments = ['subset', variable + ':' + escape_colons(filename),
                     'x=[%s,%s],y=[%s,%s]' % (lon_min, lon_max, lat_min, lat_max), '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min, 'lat', 'lon')
        self.check_output_contains_variables(self.OUTPUT_FILENAME, [variable])

    def test_GIVEN_single_variable_as_var_name_in_gridded_file_WHEN_subset_THEN_subsetted_correctly(self):
        variable = valid_hadgem_variable
        filename = valid_hadgem_filename
        lon_min, lon_max = 0, 10
        lat_min, lat_max = 40, 60
        arguments = ['subset', variable + ':' + escape_colons(filename),
                     'longitude=[%s,%s],latitude=[%s,%s]' % (lon_min, lon_max, lat_min, lat_max), '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min, 'lat', 'lon')
        self.check_output_contains_variables(self.OUTPUT_FILENAME, [variable])

    def test_GIVEN_multiple_variables_in_ungridded_file_WHEN_subset_THEN_subsetted_correctly(self):
        variable1 = valid_aerosol_cci_variable
        variable2 = valid_aerosol_cci_variable_2
        filename = valid_aerosol_cci_filename
        lon_min, lon_max = -10, 10
        lat_min, lat_max = 40, 60
        arguments = ['subset', variable1 + ',' + variable2 + ':' + escape_colons(filename),
                     'x=[%s,%s],y=[%s,%s]' % (lon_min, lon_max, lat_min, lat_max), '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, [variable1, variable2])

    def test_GIVEN_multiple_variables_in_gridded_file_WHEN_subset_THEN_subsetted_correctly(self):
        variable1 = valid_echamham_variable_1
        variable2 = valid_echamham_variable_2
        filename = valid_echamham_filename
        lon_min, lon_max = 0, 10
        lat_min, lat_max = 40, 60
        arguments = ['subset', variable1 + ',' + variable2 + ':' + escape_colons(filename),
                     'x=[%s,%s],y=[%s,%s]' % (lon_min, lon_max, lat_min, lat_max), '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min, 'lat', 'lon')
        self.check_output_contains_variables(self.OUTPUT_FILENAME, [variable1, variable2])

    def test_GIVEN_variables_specified_by_wildcard_WHEN_subset_THEN_subsetted_correctly(self):
        variable1 = 'surface_albedo???'
        variable2 = 'AOD*'
        filename = valid_aerosol_cci_filename
        lon_min, lon_max = -10, 10
        lat_min, lat_max = 40, 60
        arguments = ['subset', variable1 + ',' + variable2 + ':' + escape_colons(filename),
                     'x=[%s,%s],y=[%s,%s]' % (lon_min, lon_max, lat_min, lat_max), '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, ['AOD550', 'AOD870', 'surface_albedo550',
                                                                              'surface_albedo670', 'surface_albedo870'])


class TestTemporalSubsetAllProductsNamedVariables(BaseIntegrationTest):
    def do_subset(self, filename, time_min, time_max, variable):
        arguments = ['subset', variable + ':' + escape_colons(filename), 't=[%s,%s]' % (time_min, time_max), '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)

    def check_temporal_subsetting(self, t_min, t_max):
        import datetime
        from cf_units import Unit
        self.ds = Dataset(self.OUTPUT_FILENAME)
        cis_standard = datetime.datetime(1600, 1, 1, 0, 0, 0)
        time = self.ds.variables['time']
        datetime_min = datetime.datetime.strptime(t_min, "%Y-%m-%dT%H:%M:%S")
        datetime_max = datetime.datetime.strptime(t_max, "%Y-%m-%dT%H:%M:%S")
        # Expand the search by a second either way to avoid rounding problems
        datetime_min -= datetime.timedelta(seconds=1.5)
        datetime_max += datetime.timedelta(seconds=1.5)
        time_vals = convert_time_since_to_std_time(time[:], Unit(time.units))
        for time_val in time_vals:
            delta = datetime.timedelta(days=time_val)
            datetime_value = cis_standard + delta
            assert_that(datetime_value, greater_than_or_equal_to(datetime_min))
            assert_that(datetime_value, less_than_or_equal_to(datetime_max))

    @raises(NoDataInSubsetError)
    def test_subset_Cloud_CCI(self):
        # Takes 4s
        variable = 'time,lon,lat,satellite_zenith_view_no1,solar_zenith_view_no1'
        filename = valid_cloud_cci_filename
        time_min, time_max = '2008-07', '2008-08'
        # This is a single timestamp so the best we can do is exclude it and confirm no data is returned.
        self.do_subset(filename, time_min, time_max, variable)

    def test_subset_Aerosol_CCI(self):
        # Takes 2s
        variable = 'lat,lon,time,AOD550,AOD870,pixel_number'
        filename = valid_aerosol_cci_filename
        time_min, time_max = '2008-06-12T10:15:00', '2008-06-12T10:35:00'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_subset_NCAR_RAF(self):
        # Takes 4s
        filename = valid_NCAR_NetCDF_RAF_filename
        variable = "LATC,LONC,GGALTC,Time,PSXC,WSC,ATX,ATHR2,CONCD_LWI"
        time_min, time_max = '2009-01-14T20:15:00', '2009-01-15T02:45:00'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_subset_NCAR_RAF_with_named_time_variable(self):
        # Takes 4s
        filename = valid_NCAR_NetCDF_RAF_filename
        variable = "LATC,LONC,GGALTC,Time,PSXC,WSC,ATX,ATHR2,CONCD_LWI"
        time_min, time_max = '2009-01-14T20:15:00', '2009-01-15T02:45:00'
        arguments = ['subset', variable + ':' + escape_colons(filename), 'time=[%s,%s]' % (time_min, time_max), '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_subset_GASSP(self):
        # Takes 1.3s
        filename = valid_GASSP_aeroplane_filename
        variable = ",".join(valid_GASSP_aeroplane_vars)
        time_min, time_max = '2006-09-27T19:15:00', '2006-09-27T20:45:00'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_subset_GASSP_aux_coord(self):
        # Takes 1.3s
        filename = cis_test_files['GASSP_aux_coord'].master_filename
        variable = ",".join(cis_test_files['GASSP_aux_coord'].all_variable_names)
        time_min, time_max = '1995-11-08T17:00:00', '1995-11-08T20:00:00'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_subset_GASSP_not_entirely_valid(self):
        # see issue JASCIS-145
        filename = valid_GASSP_not_entirely_correct_filename
        variable = valid_GASSP_not_entirely_correct_variable
        time_min, time_max = '1993-10-27T00:00:00', '1993-11-27T00:00:00'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_subset_Caliop_L1(self):
        # Takes 1470s
        variable = 'Perpendicular_Attenuated_Backscatter_532,Attenuated_Backscatter_1064,' \
                   'Total_Attenuated_Backscatter_532'
        filename = valid_caliop_l1_filename
        time_min, time_max = '2009-12-31T23:40:00', '2010-01-01T00:17:17'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_subset_Aeronet(self):
        # Takes 30s
        variable = ','.join(all_aeronet_variables[0:5])
        filename = valid_aeronet_filename
        time_min, time_max = '2003-09-24T07:00:00', '2003-11-04T07:00:00'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_subset_MODIS_L2(self):
        # Takes 16s
        variable = 'Solar_Zenith,Latitude,Sensor_Azimuth,Optical_Depth_Ratio_Small_Land_And_Ocean,Sensor_Zenith,' \
                   'Scan_Start_Time,Image_Optical_Depth_Land_And_Ocean,Cloud_Fraction_Land,' \
                   'Number_Pixels_Used_Ocean,Longitude'
        filename = valid_modis_l2_filename
        time_min, time_max = '2010-01-01T22:55:19', '2010-01-01T22:58:44'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    @raises(CoordinateNotFoundError)
    def test_subset_MODIS_L3(self):
        # Takes 5s
        variable = 'Optical_Depth_Ratio_Small_Land_And_Ocean_Std_Deviation_Mean,Solar_Zenith_Std_Deviation_Mean,' \
                   'Solar_Azimuth_Std_Deviation_Mean,Optical_Depth_Ratio_Small_Land_And_Ocean_Pixel_Counts,' \
                   'Optical_Depth_Ratio_Small_Land_QA_Std_Deviation_Mean'
        filename = valid_modis_l3_filename
        time_min, time_max = '2010-01-13T00:00:01', '2010-01-13T00:01:44'
        # This is a single scalar timestamp so trying to subset on time will result in CoordinateNotFound (as we only
        # look for extended coordinates). We may want to revisit this if there is some good reason for subsetting scalar
        # coordinates...
        self.do_subset(filename, time_min, time_max, variable)

    @skip_pyhdf
    def test_subset_CloudSatPRECIP(self):
        # Takes 17s
        variable = 'Profile_time,Latitude,Longitude,DEM_elevation,Data_quality'
        filename = valid_cloudsat_PRECIP_file
        time_min, time_max = '2008-02-14T00:57:36', '2008-02-14T02:09:36'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    @skip_pyhdf
    def test_subset_Caliop_L2(self):
        # Takes 25s
        variable = 'Perpendicular_Backscatter_Coefficient_532,' \
                   'Perpendicular_Backscatter_Coefficient_Uncertainty_532,Pressure'
        filename = valid_caliop_l2_filename
        time_min, time_max = '2009-12-31T23:42:43', '2010-01-01T00:17:17'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    @skip_pyhdf
    def test_subset_CloudSatRVOD(self):

        variable = "RVOD_liq_water_content,RVOD_ice_water_path"
        filename = valid_cloudsat_RVOD_file
        time_min, time_max = '2007-06-29T13:12:00', '2007-06-29T14:29:00'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_subset_cis_gridded(self):
        # Takes 1s
        variable = 'TAU_2D_550nm'
        filename = valid_cis_gridded_output_filename
        time_min, time_max = '2007-06-04T10:18:37', '2007-06-28T10:19:47'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_subset_cis_ungridded(self):
        # Takes 1s
        variable = 'AOD550,AOD870,latitude,time'
        filename = valid_cis_ungridded_output_filename
        time_min, time_max = '2008-06-12T10:18:37', '2008-06-12T10:19:47'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_subset_netCDF_gridded_HadGem(self):
        # Takes 1s
        variable = 'od550aer'
        filename = valid_hadgem_filename
        time_min, time_max = '2007-06-02T10:18:37', '2007-06-12T10:19:47'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))

    def test_subset_ASCII(self):
        variable = 'value'
        filename = valid_ascii_filename
        time_min, time_max = '2012-08-23T15:32:03', '2012-08-28T00:00:00'
        self.do_subset(filename, time_min, time_max, variable)
        self.check_temporal_subsetting(time_min, time_max)
        self.check_output_contains_variables(self.OUTPUT_FILENAME, variable.split(','))


class TestSpatialSubsetAllProductsAllValidVariables(BaseIntegrationTest):
    def do_subset(self, filename, lat_max, lat_min, lon_max, lon_min, variable):
        arguments = ['subset', variable + ':' + escape_colons(filename),
                     'x=[%s,%s],y=[%s,%s]' % (lon_min, lon_max, lat_min, lat_max), '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)

    def test_subset_Cloud_CCI(self):
        # (All variables takes 30 mins)
        variable = '*'  # Would like to run this but it takes up a lot of memory on Jenkins.
        variable = 'time,lon,lat,satellite_zenith_view_no1,solar_zenith_view_no1'
        filename = valid_cloud_cci_filename
        lon_min, lon_max = 84, 99
        lat_min, lat_max = -6, 6
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    def test_subset_Aerosol_CCI(self):
        # Takes 97s
        variable = '*'
        filename = valid_aerosol_cci_filename
        lon_min, lon_max = -15, 5
        lat_min, lat_max = -30, 45
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    def test_subset_NCAR_RAF(self):
        # Takes 170s
        variable = 'ATX,PITCH'
        filename = valid_NCAR_NetCDF_RAF_filename
        lon_min, lon_max = -160, -157
        lat_min, lat_max = 30, 50
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    @skip_pyhdf
    def test_subset_Caliop_L1(self):
        # Takes 473s
        variable = '*'
        filename = valid_caliop_l1_filename
        lon_min, lon_max = 0, 60
        lat_min, lat_max = -30, 30
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    def test_subset_Aeronet(self):
        # Takes 60s
        variable = 'AOT_*'
        filename = valid_aeronet_filename
        lon_min, lon_max = -1.5, 1.4
        lat_min, lat_max = 15, 15.5
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    @skip_pyhdf
    def test_subset_MODIS_L2(self):
        # Takes 35s
        variable = '*'
        filename = valid_modis_l2_filename
        lon_min, lon_max = -150, 150
        lat_min, lat_max = -72, -63
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    @skip_pyhdf
    def test_subset_MODIS_L3(self):
        variable = '*'
        filename = valid_modis_l3_filename
        lon_min, lon_max = -10, 10
        lat_min, lat_max = 40, 60
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    def test_subset_netCDF_gridded_HadGem(self):
        # Takes 1s
        variable = '*'
        filename = valid_hadgem_filename
        lon_min, lon_max = 0, 120
        lat_min, lat_max = 40, 60
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min, 'lat', 'lon')

    @skip_pyhdf
    def test_subset_Caliop_L2(self):
        # Takes 40s
        variable = '*'
        filename = valid_caliop_l2_filename
        lon_min, lon_max = 0, 60
        lat_min, lat_max = -30, 30
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    @skip_pyhdf
    def test_subset_CloudSatPRECIP(self):
        # Takes 100s
        variable = '*'
        filename = valid_cloudsat_PRECIP_file
        lon_min, lon_max = -10, 10
        lat_min, lat_max = 40, 60
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    def test_subset_cis_gridded(self):
        # Takes 1s
        variable = '*'
        filename = valid_cis_gridded_output_filename
        lon_min, lon_max = 0, 10
        lat_min, lat_max = 40, 60
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min, 'lat', 'lon')

    def test_subset_cis_ungridded(self):
        # Takes 1s
        variable = '*'
        filename = valid_cis_ungridded_output_filename
        lon_min, lon_max = 1, 3
        lat_min, lat_max = 41, 42
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    @skip_pyhdf
    def test_subset_CloudSatRVOD(self):
        # 257s exit code 137
        variable = '*'  # Gets killed by Jenkins
        variable = "RVOD_liq_water_path,RVOD_ice_water_path"
        filename = valid_cloudsat_RVOD_file
        lon_min, lon_max = -10, 10
        lat_min, lat_max = 40, 60
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    def test_subset_GASP(self):
        # 257s exit code 137
        variable = '*'
        filename = valid_GASSP_aeroplane_filename
        lon_min, lon_max = -94, 95
        lat_min, lat_max = 30, 31
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    def test_subset_GASSP_aux_coord(self):
        # Takes 1.3s
        filename = cis_test_files['GASSP_aux_coord'].master_filename
        variable = ",".join(cis_test_files['GASSP_aux_coord'].all_variable_names)
        lon_min, lon_max = -156, -155
        lat_min, lat_max = 5, 10
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)

    def test_subset_ASCII(self):
        variable = '*'
        filename = valid_ascii_filename
        lon_min, lon_max = -10, 10
        lat_min, lat_max = 1, 6
        self.do_subset(filename, lat_max, lat_min, lon_max, lon_min, variable)
        self.check_latlon_subsetting(lat_max, lat_min, lon_max, lon_min)


class TestEmptySubsets(BaseIntegrationTest):

    def do_subset(self, filename, variable, alt_bounds='', pres_bounds=''):
        # Join the bounds with a comma if they are both specified
        joint_bounds = ','.join([alt_bounds, pres_bounds]) if alt_bounds and pres_bounds else alt_bounds or pres_bounds
        arguments = ['subset', variable + ':' + escape_colons(filename), joint_bounds, '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)

    @raises(NoDataInSubsetError)
    def test_empty_subset_ungridded_vertical(self):
        # Takes 170s
        variable = valid_NCAR_NetCDF_RAF_variable
        filename = valid_NCAR_NetCDF_RAF_filename
        alt_min, alt_max = 15000, 20000
        self.do_subset(filename, variable, alt_bounds='z=[{},{}]'.format(alt_min, alt_max))
        self.check_alt_subsetting(alt_max, alt_min)

    #More...


class TestVerticalSubsetAllProducts(BaseIntegrationTest):
    def do_subset(self, filename, variable, alt_bounds='', pres_bounds=''):
        # Join the bounds with a comma if they are both specified
        joint_bounds = ','.join([alt_bounds, pres_bounds]) if alt_bounds and pres_bounds else alt_bounds or pres_bounds
        arguments = ['subset', variable + ':' + escape_colons(filename), joint_bounds, '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)

    def test_subset_NCAR_RAF_alt(self):
        # Takes 170s
        variable = valid_NCAR_NetCDF_RAF_variable
        filename = valid_NCAR_NetCDF_RAF_filename
        alt_min, alt_max = 1000, 2000
        self.do_subset(filename, variable, alt_bounds='z=[{},{}]'.format(alt_min, alt_max))
        self.check_alt_subsetting(alt_max, alt_min)

    def test_subset_NCAR_RAF_pres(self):
        # Takes 170s
        variable = valid_NCAR_NetCDF_RAF_variable
        filename = valid_NCAR_NetCDF_RAF_filename
        pres_min, pres_max = 1000, 2000
        self.do_subset(filename, variable, alt_bounds='p=[{},{}]'.format(pres_min, pres_max))
        self.check_pres_subsetting(pres_max, pres_min)

    def test_subset_NCAR_RAF_pres_order_doesnt_matter(self):
        # Takes 170s
        variable = valid_NCAR_NetCDF_RAF_variable
        filename = valid_NCAR_NetCDF_RAF_filename
        pres_min, pres_max = 1000, 2000
        self.do_subset(filename, variable, alt_bounds='p=[{},{}]'.format(pres_max, pres_min))
        self.check_pres_subsetting(pres_max, pres_min)

    @skip_pyhdf
    def test_subset_Caliop_L1(self):
        # Takes 473s
        variable = ','.join(valid_caliop_l1_variables)
        filename = valid_caliop_l1_filename
        alt_min, alt_max = 1000, 2000
        self.do_subset(filename, variable, alt_bounds='z=[{},{}]'.format(alt_min, alt_max))
        self.check_alt_subsetting(alt_max, alt_min)

    def test_subset_hybrid_pressure_model_field(self):
        """
        In the case of subsetting hybrid height/pressure fields using the axis name CIS (iris) will choose the vertical
        *dimension* coordinate, whether that is model level number, hybrid sigma, or something else
        """
        # Takes 1s
        variable = valid_hybrid_pressure_variable
        filename = valid_hybrid_pressure_filename
        pres_min, pres_max = 0.95, 0.96
        self.do_subset(filename, variable, alt_bounds='z=[{},{}]'.format(pres_min, pres_max))
        self.check_pres_subsetting(pres_max, pres_min, pres_name='lev')

    @raises(CoordinateNotFoundError)
    def test_subset_hybrid_pressure_model_field_by_variable(self):
        """
        In the case of hybrid height/pressure fields specifying the hybrid variable directly will result in a coordinate
          not found error
        """
        # Takes 1s
        variable = valid_hybrid_pressure_variable
        filename = valid_hybrid_pressure_filename
        pres_min, pres_max = 1000, 2000
        self.do_subset(filename, variable, alt_bounds='air_pressure=[{},{}]'.format(pres_min, pres_max))
        self.check_pres_subsetting(pres_max, pres_min)

    @skip_pyhdf
    def test_subset_Caliop_L2(self):
        # Takes 40s
        variable = ','.join(valid_caliop_l2_variables)
        filename = valid_caliop_l2_filename
        alt_min, alt_max = 1000, 2000
        self.do_subset(filename, variable, alt_bounds='z=[{},{}]'.format(alt_min, alt_max))
        self.check_alt_subsetting(alt_max, alt_min)

    @skip_pyhdf
    def test_subset_CloudSatRVOD_alt(self):
        # 257s exit code 137
        variable = "RVOD_liq_water_content,RVOD_ice_water_path"
        filename = valid_cloudsat_RVOD_file
        alt_min, alt_max = 0, 2000
        self.do_subset(filename, variable, alt_bounds='z=[{},{}]'.format(alt_min, alt_max))
        self.check_alt_subsetting(alt_max, alt_min)

    def test_subset_GASSP(self):
        # 257s exit code 137
        variable = '*'
        filename = valid_GASSP_aeroplane_filename
        alt_min, alt_max = 0, 2000
        self.do_subset(filename, variable, alt_bounds='z=[{},{}]'.format(alt_min, alt_max))
        self.check_alt_subsetting(alt_max, alt_min)

    @raises(CoordinateNotFoundError)
    def test_subset_GASSP_pres(self):
        # 257s exit code 137
        variable = '*'
        filename = valid_GASSP_aeroplane_filename
        pres_min, pres_max = 2000, 200
        self.do_subset(filename, variable, alt_bounds='p=[{},{}]'.format(pres_min, pres_max))

    def test_subset_ASCII(self):
        variable = '*'
        filename = valid_ascii_filename
        alt_min, alt_max = 0, 12
        self.do_subset(filename, variable, alt_bounds='z=[{},{}]'.format(alt_min, alt_max))
        self.check_alt_subsetting(alt_max, alt_min)


class TestAuxSubset(BaseIntegrationTest):

    def check_aux_subsetting(self, aux_min, aux_max, var):
        self.ds = Dataset(self.OUTPUT_FILENAME)
        alt = self.ds.variables[var][:]
        assert_that(alt.min(), greater_than_or_equal_to(aux_min))
        assert_that(alt.max(), less_than_or_equal_to(aux_max))
        self.ds.close()

    def do_subset(self, filename, variable, aux_bounds):
        # Join the bounds with a comma if they are both specified
        arguments = ['subset', variable + ':' + escape_colons(filename), aux_bounds, '-o', self.OUTPUT_FILENAME]
        main_arguments = parse_args(arguments)
        subset_cmd(main_arguments)

    def test_subset_GASSP_aux_coord(self):
        # Takes 1.3s
        filename = cis_test_files['GASSP_aux_coord'].master_filename
        variable = cis_test_files['GASSP_aux_coord'].data_variable_name
        dp_min, dp_max = 0.05, 0.8
        self.do_subset(filename, variable, 'DP_MID=[{min},{max}]'.format(min=dp_min, max=dp_max))
        self.check_aux_subsetting(dp_min, dp_max, 'DP_MID')



#  tests for cis-1.7.3-py36ha15d459_1 (this is a generated file);
print('===== testing package: cis-1.7.3-py36ha15d459_1 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
from cis.test.runner import run
import os

# Disable multi proccesing completely - the timeout is breaking builds on OS X as well now...
np=0
# Run the tests
run('cis.test.unit', n_processors=np)

# If the test data directory is specified, run the integration tests as well. We disable parallel processing
# of these though as some have side affects, you also have to specify a time-out which would have to be very large
if os.environ.get("CIS_DATA_HOME", None):
    run('cis.test.integration', n_processors=0)
#  --- run_test.py (end) ---

print('===== cis-1.7.3-py36ha15d459_1 OK =====');
print("import: 'cis'")
import cis




"""
A basic histogram plot
"""
from cis.plotting.genericplot import GenericPlot


class Histogram(GenericPlot):
    valid_histogram_styles = ["bar", "step", "stepfilled"]

    def __init__(self, packed_data, xbins=10, *args, **kwargs):
        """
        :param xbins: The number of bins to use on the xaxis
        """
        super(Histogram, self).__init__(packed_data, *args, **kwargs)
        self.xbins = xbins
        #
        self.xlabel = packed_data.name()
        self.ylabel = "Frequency"

    def __call__(self, ax):
        from numpy.ma import MaskedArray

        self.mplkwargs["bins"] = self.xbins

        if self.itemstyle:
            if self.itemstyle in self.valid_histogram_styles:
                self.mplkwargs["histtype"] = self.itemstyle
            else:
                from cis.exceptions import InvalidHistogramStyleError
                raise InvalidHistogramStyleError(
                    "'" + self.itemstyle + "' is not a valid histogram style, please use one of: " + str(
                        self.valid_histogram_styles))

        if self.color:
            self.mplkwargs["color"] = self.color

        if isinstance(self.data, MaskedArray):
            data = self.data.compressed()
        else:
            data = self.data.flatten()

        ax.hist(data, *self.mplargs, **self.mplkwargs)

        super(Histogram, self).__call__(ax)

    def unpack_data_items(self, packed_data_items):
        self.data = packed_data_items.data



"""
Module for writing data to NetCDF files
"""
from netCDF4 import Dataset
import logging

types = {'int8': 'i1',
         'int16': "i2",
         'int32': "i4",
         'int64': "i8",
         'uint8': 'u1',
         'uint16': "u2",
         'uint32': "u4",
         'uint64': "u8",
         'float32': "f4",
         'float64': "f8"}

index_name = 'obs'


def __add_metadata(var, data):
    if data.standard_name:
        var.standard_name = data.standard_name
    if data.units:
        var.units = str(data.units)
    if data.long_name:
        var.long_name = data.long_name
    if data.metadata.missing_value:
        var.missing_value = data.metadata.missing_value
    if hasattr(data.units, 'calendar') and data.units.calendar:
        var.calendar = data.units.calendar
    if data.metadata.history:
        var.history = data.metadata.history
    for name, value in data.attributes.items():
        setattr(var, name, value)
    for name, value in data.metadata.misc.items():
        if name not in var.ncattrs():
            try:
                setattr(var, name, value)
            except ValueError:
                logging.warning("Invalid value ({}) for {} attribute. "
                                "Attribute not saved to {}.".format(value, name, var.name))
    return var


def __get_missing_value(coord):
    f = coord.metadata.missing_value
    if not f and f != 0:
        f = None
    return f


def sizeof_fmt(num, suffix='B'):
    """
    Return a human readable size from an integer number of bytes

    From http://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
    :param int num: Number of bytes
    :param str suffix: Little or big B
    :return str: Formatted human readable size
    """
    for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def __check_disk_space(filepath, data):
    """
    Warn if there is insufficient disk space to write the data to the filapath - if the OS supports it
    :param str filepath: Path of file to write
    :param ndarray data: Numpy array of data to write
    :return: None
    """
    try:
        from os import statvfs
        stats = statvfs(filepath)
    except (ImportError, OSError):
        logging.debug("Unable to determine free disk space on this OS.")
    else:
        # available space is the number of available blocks times the fundamental block size
        available = stats.f_bavail * stats.f_frsize
        if available < data.nbytes:
            logging.warning("Free disk space at {path} is {free}, but the array being saved is {size}."
                            .format(path=filepath, free=sizeof_fmt(available), size=sizeof_fmt(data.data.nbytes)))


def __create_variable(nc_file, data, prefer_standard_name=False):
    """Creates and writes a variable to a netCDF file.
    :param nc_file: netCDF file to which to write
    :param data: LazyData for variable to write
    :param prefer_standard_name: if True, use the standard name of the variable if defined,
           otherwise use the variable name
    :return: created netCDF variable
    """
    from cis.exceptions import InconsistentDimensionsError

    name = None
    if (data.metadata._name is not None) and (len(data.metadata._name) > 0):
        name = data.metadata._name
    if (name is None) or prefer_standard_name:
        if (data.metadata.standard_name is not None) and (len(data.metadata.standard_name) > 0):
            name = data.metadata.standard_name
    out_type = types[str(data.data.dtype)]
    logging.info("Creating variable: {name}({index}) {type}".format(name=name, index=index_name, type=out_type))
    if name not in nc_file.variables:
        # Generate a warning if we have insufficient disk space
        __check_disk_space(nc_file.filepath(), data.data)
        var = nc_file.createVariable(name, datatype=out_type, dimensions=index_name,
                                     fill_value=__get_missing_value(data))
        var = __add_metadata(var, data)
        try:
            var[:] = data.data.flatten()
        except IndexError as e:
            raise InconsistentDimensionsError(str(e) + "\nInconsistent dimensions in output file, unable to write "
                                                       "{} to file (it's shape is {}).".format(data.name(), data.shape))
        except:
            logging.error("Error writing data to disk.")
            raise
        return var
    else:
        return nc_file.variables[name]


def write(data_object, filename):
    """

    :param data_object:
    :param filename:
    :return:
    """
    write_coordinate_list(data_object.coords(), filename)
    add_data_to_file(data_object, filename)


def write_coordinates(coords, filename):
    """Writes coordinates to a netCDF file.

    :param coords: UngriddedData or UngriddedCoordinates object for which the coordinates are to be written
    :param filename: file to which to write
    """
    coord_list = coords.coords()
    write_coordinate_list(coord_list, filename)


def write_coordinate_list(coord_list, filename):
    """Writes coordinates to a netCDF file.

    :param coord_list: list of Coord objects
    :param filename: file to which to write
    """
    netcdf_file = Dataset(filename, 'w', format="NETCDF4")
    try:
        length = len(coord_list[0].data.flatten())
    except AttributeError:
        length = len(coord_list[0].points.flatten())
    _ = netcdf_file.createDimension(index_name, length)
    for coord in coord_list:
        __create_variable(netcdf_file, coord, prefer_standard_name=True)
    netcdf_file.close()


def add_data_to_file(data_object, filename):
    """

    :param data_object:
    :param filename:
    :return:
    """
    from cis import __version__
    netcdf_file = Dataset(filename, 'a', format="NETCDF4")
    var = __create_variable(netcdf_file, data_object, prefer_standard_name=False)
    netcdf_file.source = "CIS" + __version__
    netcdf_file.close()



"""
Module containing NetCDF file reading functions
"""
from cis.exceptions import InvalidVariableError
from cis.utils import listify
import logging

def get_netcdf_file_attributes(filename):
    """
    Get all the global attributes from a NetCDF file

    :param filename: The filename of the file to get the variables from
    :return: a dictionary of attributes and their values
    """

    from netCDF4 import Dataset
    try:
        f = Dataset(filename)
    except RuntimeError as e:
        raise IOError(e)

    return f.__dict__


def get_netcdf_file_variables(filename, exclude_coords=False):
    """
    Get all the variables contained in a NetCDF file. Variables in NetCDF4 Hierarchical groups are returned with their
    fully qualified variable name in the form ``<group1>/<group2....>/<variable_name>``,
    e.g.``AVHRR/Ch4CentralWavenumber``.

    :param filename: The filename of the file to get the variables from
    :param exclude_coords: Exclude coordinate variables if True
    :return: An OrderedDict containing {variable_name: NetCDF Variable instance}
    """
    from netCDF4 import Dataset
    try:
        f = Dataset(filename)
    except RuntimeError as e:
        raise IOError(e)
    # Include NetCDF4 hierarchical groups too:
    variables = _get_all_fully_qualified_variables(f)
    if exclude_coords:
        for var in f.dimensions:
            try:
                del variables[var]
            except KeyError:
                pass
    return variables


def _get_all_fully_qualified_variables(dataset):
    """
    List all variables in a file.
    Variable names may be fully qualified NetCDF4 Hierarchical group variables in the form
    <group1>/<group2....>/<variable_name>, e.g. 'AVHRR/Ch4CentralWavenumber'.
    :param dataset: Dataset to get variables for
    :return: Dictionary of variable_name : netCDF4 Variable instance
    """

    def get_variables_for_group(group, var_dict, previous_groups=None):
        """
        Recursively get Variables for this group and all subgroups,
        adding them to a dictionary.
        :param group: Group to start from (typically root group, i.e. netCDF4.Dataset)
        :param var_dict: Dictionary to add Variable instances to
        :param previous_groups: List of all groups up to this point  in the recursion
        """
        if not previous_groups:
            previous_groups = []
        for group_key, group in group.groups.items():
            current_groups = previous_groups + [group_key]
            for var_key, var in group.variables.items():
                path = current_groups + [var_key]
                var_dict["/".join(path)] = var
            get_variables_for_group(group, var_dict, current_groups)

    all_vars = dataset.variables
    get_variables_for_group(dataset, all_vars)
    return all_vars


def remove_variables_with_non_spatiotemporal_dimensions(variables, spatiotemporal_var_names):
    """
    Remove from a list of netCDF variables any which have dimensionality which is not in an approved list
    of valid spatial or temporal dimensions (e.g. sensor number, pseudo dimensions). CIS currently does not
    support variables with this dimensionality and will fail if they are used.

    :param variables: Dictionary of netCDF variable names : Variable objects. Variable names may be fully qualified
      NetCDF4 Hierarchical group variables in the form ``<group1>/<group2....>/<variable_name>``,
      e.g. ``AVHRR/Ch4CentralWavenumber``.
    :param spatiotemporal_var_names: List of valid spatiotemporal dimensions.
    :return: None
    """
    if spatiotemporal_var_names is not None:
        for var in list(variables.keys()):
            for dim in variables[var].dimensions:
                if dim not in spatiotemporal_var_names:
                    del variables[var]
                    break


def read_many_files(filenames, usr_variables, dim=None):
    """
    Reads a single Variable from many NetCDF files. This method uses the netCDF4 MFDataset class and so is NOT
    suitable for NetCDF4 datasets (only 'CLASSIC' netcdf).

    :param filenames: A list of NetCDF filenames to read, or a string with wildcards.
    :param usr_variables: A list of variable (dataset) names to read from the files.
      The names must appear exactly as in in the NetCDF file.
    :param dim: The name of the dimension on which to aggregate the data. None is the default
      which tries to aggregate over the unlimited dimension
    :return: A list of variable instances constructed from all of the input files
    """
    from netCDF4 import MFDataset
    from cis.exceptions import InvalidVariableError

    usr_variables = listify(usr_variables)

    try:
        datafile = MFDataset(filenames, aggdim=dim)
    except RuntimeError as e:
        raise IOError(e)

    data = {}
    for variable in usr_variables:
        # Get data.
        try:
            data[variable] = datafile.variables[variable]
        except:
            raise InvalidVariableError('Variable {} not found in file {}.'.format(variable, filenames))

    return data


def read_many_files_individually(filenames, usr_variables):
    """
    Read multiple Variables from many NetCDF files manually - i.e. not with MFDataset as this doesn't always work,
    in particular for NetCDF4 files.

    :param filenames: A list of NetCDF filenames to read, or a string with wildcards.
    :param usr_variables: A list of variable (dataset) names to read from the files. The names must appear exactly as
      in in the NetCDF file. Variable names may be fully qualified NetCDF4 Hierarchical group variables in the form
      ``<group1>/<group2....>/<variable_name>``, e.g. ``AVHRR/Ch4CentralWavenumber``.
    :return: A dictionary of lists of variable instances constructed from all of the input files with the fully
      qualified variable name as the key
    """
    from cis.utils import add_element_to_list_in_dict

    usr_variables = listify(usr_variables)

    var_data = {}

    for filename in filenames:

        var_dict = read(filename, usr_variables)
        for var in list(var_dict.keys()):
            add_element_to_list_in_dict(var_data, var, var_dict[var])

    return var_data


def read(filename, usr_variables):
    """
    Reads a Variable from a NetCDF file

    :param filename: The name (with path) of the NetCDF file to read.
    :param usr_variables: A variable (dataset) name to read from the files. The name must appear exactly as in in the
      NetCDF file. Variable names may be fully qualified NetCDF4 Hierarchical group variables in the form
      ``<group1>/<group2....>/<variable_name>``, e.g. ``AVHRR/Ch4CentralWavenumber``.
    :return: A Variable instance constructed from  the input file
    """
    from netCDF4 import Dataset

    usr_variables = listify(usr_variables)

    try:
        datafile = Dataset(filename)
    except RuntimeError as e:
        raise IOError(str(e))

    data = {}
    for full_variable in usr_variables:
        # Split the fully qualified variable (group/variable) into group and variable
        parts = full_variable.split("/")
        groups = parts[:-1]
        variable = parts[-1]
        current_group = datafile
        for group in groups:
            current_group = current_group.groups[group]
        try:
            data[full_variable] = current_group.variables[variable]
        except:
            raise InvalidVariableError(full_variable + ' could not be found in ' + filename)

    return data


def get_metadata(var):
    """
    Retrieves all metadata

    :param var: the Variable to read metadata from
    :return: A metadata object
    """
    from cis.data_io.ungridded_data import Metadata
    from cis.utils import set_standard_name_if_valid

    missing_value = find_missing_value(var)

    standard_attributes = ['standard_name', 'long_name', 'history', 'units']

    attrs = {attr: getattr(var, attr, "") for attr in standard_attributes}

    shape = getattr(var, "shape", None)
    if shape is None:
        try:
            shape = (var._recLen[0],)
        except AttributeError:
            shape = ()

    metadata = Metadata(
        var._name,
        long_name=attrs['long_name'],
        units=attrs['units'],
        missing_value=missing_value,
        shape=shape,
        history=attrs['history'],
        misc={k: getattr(var, k) for k in var.ncattrs() if k not in standard_attributes})

    # Only set the standard name if it's CF compliant
    set_standard_name_if_valid(metadata, attrs['standard_name'])

    return metadata


def find_missing_value(var):
    """
    Get the missing / fill value of the variable

    :param var: NetCDF Variable instance
    :return: missing / fill value
    """
    try:
        missing_value = var.missing_value
    except AttributeError:
        try:
            missing_value = var._FillValue
        except AttributeError:
            missing_value = None
    return missing_value


def get_data(var):
    """
    Reads raw data from a NetCDF.Variable instance. Also applies CF-compliant valid max, min and ranges.

    :param var: The specific Variable instance to read
    :return:  A numpy maskedarray. Missing values are False in the mask.
    """
    import numpy as np
    import logging
    # Turn off scaling and masking as we're a bit more lenient about the type of valid min/max.
    var.set_auto_maskandscale(False)
    # This will still automatically return a masked array based on _FillValue and missing_value
    data = var[:]

    if hasattr(var, 'valid_max'):
        try:
            v_max = np.array(var.valid_max, var.dtype)
        except ValueError:
            logging.warning("Unable to parse valid_max metadata for {}. Not applying mask.".format(var._name))
        else:
            logging.debug("Masking all values > {}.".format(v_max))
            data = np.ma.masked_greater(data, v_max)

    if hasattr(var, 'valid_min'):
        try:
            v_min = np.array(var.valid_min, var.dtype)
        except ValueError:
            logging.warning("Unable to parse valid_min metadata for {}. Not applying mask.".format(var._name))
        else:
            logging.debug("Masking all values < {}.".format(v_min))
            data = np.ma.masked_less(data, v_min)

    if hasattr(var, 'valid_range'):
        try:
            data = np.ma.masked_outside(data, *var.valid_range)
            logging.debug("Masking all values {} > v > {}.".format(*var.valid_range))
        except (ValueError, TypeError):
            logging.warning("Unable to parse valid_range metadata for {}. Not applying mask.".format(var._name))

    # Now apply any scaling
    data = apply_offset_and_scaling(data, getattr(var, 'add_offset', None), getattr(var, 'scale_factor', None))

    return data


def apply_offset_and_scaling(data, add_offset=None, scale_factor=None):
    """
    Apply a standard offset and scaling to the data. This is deliberately very similar to the python-NetCDF4
    implementation as it is anticipated that we can remove it if/when that library implements valid_range masking.

    :param ndarray data: Data to scale
    :param float add_offset:
    :param float scale_factor:
    :return ndarray: Scaled data
    """
    if scale_factor is not None and add_offset is not None and \
            (add_offset != 0.0 or scale_factor != 1.0):
        data = data * scale_factor + add_offset
        logging.debug("Applying 'data = data * {scale} + {offset}' transformation to data.".format(scale=scale_factor,
                                                                                                   offset=add_offset))
    elif scale_factor is not None and scale_factor != 1.0:
        data *= scale_factor
        logging.debug("Applying 'data *= {scale}' transformation to data.".format(scale=scale_factor))
    elif add_offset is not None and add_offset != 0.0:
        data += add_offset
        logging.debug("Applying 'data += {offset}' transformation to data.".format(offset=add_offset))
    return data



"""
Tests the cis.col.Collocate class
"""
import unittest
from nose.tools import assert_raises


class TestCollocate(unittest.TestCase):
    def test_get_kernel(self):
        from cis.collocation.col_implementations import moments, nn_pressure, mean
        from cis.collocation.col import get_kernel

        assert isinstance(get_kernel('moments'), moments)
        assert isinstance(get_kernel(nn_pressure()), nn_pressure)
        assert isinstance(get_kernel('mean', default=moments), mean)
        assert isinstance(get_kernel('', default=moments), moments)

        with assert_raises(ValueError):
            get_kernel('foo')

        with assert_raises(ValueError):
            get_kernel('foo', default=moments)


"""Tests for ungridded_data module
"""
from unittest import TestCase, skipIf
from hamcrest import assert_that, is_, contains_inanyorder
from nose.tools import raises, assert_almost_equal, assert_raises
from cis.test.util.mock import make_regular_2d_ungridded_data, make_regular_2d_ungridded_data_with_missing_values, \
    make_regular_2d_with_time_ungridded_data

import numpy as np
from cf_units import Unit

from cis.data_io.Coord import CoordList, Coord
from cis.data_io.ungridded_data import UngriddedCoordinates, UngriddedData, Metadata, UngriddedDataList

try:
    import pandas
except ImportError:
    # Disable all these tests if pandas is not installed.
    pandas = None

skip_pandas = skipIf(pandas is None, 'Test(s) require "pandas", which is not available.')


class TestMetadata(TestCase):

    def test_unit_setter(self):
        # Standard units
        m = Metadata(units='kg m-2 s-1')
        assert m.units == Unit('kg m-2 s-1')
        # Case sensitive units
        m.units = 'hPa'
        assert m.units == Unit('hPa')
        # Units with number in
        m.units = '#/cm3'
        assert m.units == Unit('cm-3')
        # standard time since units
        m.units = 'days since 1999-01-01'
        assert m.units == Unit('days since 1999-01-01')
        # Capitalised time since
        m.units = 'Days since 1999-01-01'
        assert m.units == Unit('days since 1999-01-01')
        # Time since with a colon and capital S
        m.units = 'hours Since: 1999-01-01'
        assert m.units == Unit('hours Since 1999-01-01')


class TestUngriddedData(TestCase):

    def test_can_create_ungridded_data(self):
        ug = make_regular_2d_ungridded_data()
        assert(ug.data.size == 15)

    def test_get_all_points_returns_points(self):
        ug = make_regular_2d_ungridded_data()
        points = ug.get_all_points()
        num_points = len([p for p in points])
        assert(num_points == 15)

    def test_get_non_masked_points_returns_points(self):
        ug = make_regular_2d_ungridded_data_with_missing_values()
        points = ug.get_non_masked_points()
        num_points = len([p for p in points])
        assert(num_points == 12)

    def test_get_coordinates_points_returns_points(self):
        ug = make_regular_2d_ungridded_data_with_missing_values()
        points = ug.get_coordinates_points()
        num_points = len([p for p in points])
        assert(num_points == 15)

    def test_can_add_history(self):
        ug = make_regular_2d_ungridded_data()

        new_history = 'This is a new history entry.'
        ug.add_history(new_history)
        assert(ug.metadata.history.find(new_history) >= 0)

    def test_GIVEN_numpy_array_data_with_missing_coordinate_values_WHEN_data_THEN_missing_values_removed(self):
        x_points = np.arange(-10, 11, 5)
        y_points = np.arange(-5, 6, 5)
        y, x = np.meshgrid(y_points, x_points)
        y = np.ma.masked_array(y, np.zeros(y.shape, dtype=bool))
        y.mask[1, 2] = True

        x = Coord(x, Metadata(standard_name='latitude', units='degrees'))
        y = Coord(y, Metadata(standard_name='longitude', units='degrees'))
        coords = CoordList([x, y])

        data = np.reshape(np.arange(15) + 1.0, (5, 3))

        ug = UngriddedData(data, Metadata(), coords)
        data = ug.data.flatten()
        assert_that(len(data), is_(14))
        for coord in ug.coords():
            assert_that(len(coord.points), is_(14))

    @skip_pandas
    def test_GIVEN_ungridded_data_WHEN_call_as_data_frame_THEN_returns_valid_data_frame(self):
        from cis.test.util.mock import make_regular_4d_ungridded_data
        from datetime import datetime

        ug_data = make_regular_4d_ungridded_data()
        df = ug_data.as_data_frame()

        assert_that(df['rainfall_flux'][5] == 6)
        assert_that(df['latitude'][17] == 0)
        assert_that(df.loc[datetime(1984,8,31), 'latitude'][0] == 10)
        assert_that(df['rainfall_flux'].median() == 25.5)

    @skip_pandas
    def test_GIVEN_ungridded_data_with_no_time_WHEN_call_as_data_frame_THEN_returns_valid_data_frame(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data

        ug_data = make_regular_2d_ungridded_data()
        df = ug_data.as_data_frame()

        assert_that(df['rainfall_rate'][5] == 6)
        assert_that(df['latitude'][7] == 0)
        assert_that(df['rainfall_rate'].median() == 8.0)

    @skip_pandas
    def test_GIVEN_ungridded_data_with_missing_vals_WHEN_call_as_data_frame_THEN_returns_valid_data_frame(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data_with_missing_values

        ug_data = make_regular_2d_ungridded_data_with_missing_values()
        df = ug_data.as_data_frame()

        assert_that(df['rainfall_rate'][5] == 6)
        assert_that(df['latitude'][7] == 0)
        assert_that(df['rainfall_rate'].median() == 7.5)
        assert_that(np.isnan(df['rainfall_rate'][8]))

    def test_name_method(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data_with_missing_values
        d = make_regular_2d_ungridded_data_with_missing_values()
        # Standard name
        assert d.name() == 'rainfall_rate'
        d.standard_name = None
        # Long name
        assert d.name() == 'TOTAL RAINFALL RATE: LS+CONV KG/M2/S'
        d.long_name = ''
        # Var_name
        assert d.name() == 'rain'
        d.var_name = None
        # Default
        assert d.name() == 'unknown'
        # User specified default
        assert d.name('test') == 'test'
        # Empty default
        assert d.name('') == ''


class TestUngriddedDataLazyLoading(TestCase):

    def test_GIVEN_missing_coord_values_WHEN_data_THEN_missing_values_removed(self):
        x_points = np.arange(-10, 11, 5)
        y_points = np.arange(-5, 6, 5)
        y, x = np.meshgrid(y_points, x_points)
        y = np.ma.masked_array(y, np.zeros(y.shape, dtype=bool))
        y.mask[1, 2] = True

        x = Coord(x, Metadata(standard_name='latitude', units='degrees'))
        y = Coord(y, Metadata(standard_name='longitude', units='degrees'))
        coords = CoordList([x, y])

        data = np.reshape(np.arange(15) + 1.0, (5, 3))

        ug = UngriddedData(None, Metadata(), coords, lambda x: data)
        data = ug.data.flatten()
        assert_that(len(data), is_(14))

    def test_GIVEN_missing_coord_values_WHEN_data_flattened_THEN_missing_values_removed(self):
        x_points = np.arange(-10, 11, 5)
        y_points = np.arange(-5, 6, 5)
        y, x = np.meshgrid(y_points, x_points)
        y = np.ma.masked_array(y, np.zeros(y.shape, dtype=bool))
        y.mask[1, 2] = True

        x = Coord(x, Metadata(standard_name='latitude', units='degrees'))
        y = Coord(y, Metadata(standard_name='longitude', units='degrees'))
        coords = CoordList([x, y])

        data = np.reshape(np.arange(15) + 1.0, (5, 3))

        ug = UngriddedData(None, Metadata(), coords, lambda x: data)
        data = ug.data_flattened
        assert_that(len(data), is_(14))

    def test_GIVEN_missing_coord_values_WHEN_coords_points_THEN_missing_values_removed(self):
        x_points = np.arange(-10, 11, 5)
        y_points = np.arange(-5, 6, 5)
        y, x = np.meshgrid(y_points, x_points)
        y = np.ma.masked_array(y, np.zeros(y.shape, dtype=bool))
        y.mask[1, 2] = True

        x = Coord(x, Metadata(standard_name='latitude', units='degrees'))
        y = Coord(y, Metadata(standard_name='longitude', units='degrees'))
        coords = CoordList([x, y])

        data = np.reshape(np.arange(15) + 1.0, (5, 3))

        ug = UngriddedData(None, Metadata(), coords, lambda x: data)
        coords = ug.coords()
        for coord in coords:
            assert_that(len(coord.points), is_(14))

    def test_GIVEN_missing_coord_values_WHEN_coords_data_THEN_missing_values_removed(self):
        x_points = np.arange(-10, 11, 5)
        y_points = np.arange(-5, 6, 5)
        y, x = np.meshgrid(y_points, x_points)
        y = np.ma.masked_array(y, np.zeros(y.shape, dtype=bool))
        y.mask[1, 2] = True

        x = Coord(x, Metadata(standard_name='latitude', units='degrees'))
        y = Coord(y, Metadata(standard_name='longitude', units='degrees'))
        coords = CoordList([x, y])

        data = np.reshape(np.arange(15) + 1.0, (5, 3))

        ug = UngriddedData(None, Metadata(), coords, lambda x: data)
        coords = ug.coords()
        for coord in coords:
            assert_that(len(coord.data), is_(14))

    def test_GIVEN_missing_coord_values_WHEN_coords_flattened_THEN_missing_values_removed(self):
        x_points = np.arange(-10, 11, 5)
        y_points = np.arange(-5, 6, 5)
        y, x = np.meshgrid(y_points, x_points)
        y = np.ma.masked_array(y, np.zeros(y.shape, dtype=bool))
        y.mask[1, 2] = True

        x = Coord(x, Metadata(standard_name='latitude', units='degrees'))
        y = Coord(y, Metadata(standard_name='longitude', units='degrees'))
        coords = CoordList([x, y])

        data = np.reshape(np.arange(15) + 1.0, (5, 3))

        ug = UngriddedData(None, Metadata(), coords, lambda x: data)
        coords = ug.coords_flattened
        for coord in coords:
            if coord is not None:
                assert_that(len(coord), is_(14))


class TestUngriddedCoordinates(TestCase):

    def setUp(self):
        x_points = np.arange(-10, 11, 5)
        y_points = np.arange(-5, 6, 5)
        y, x = np.meshgrid(y_points, x_points)

        self.x = Coord(x, Metadata(standard_name='latitude', units='degrees'))
        self.y = Coord(y, Metadata(standard_name='longitude', units='degrees'))

        coords = CoordList([self.x, self.y])
        self.ug = UngriddedCoordinates(coords)

    def test_can_create_ungridded_coordinates(self):
        standard_coords = self.ug.coords().find_standard_coords()
        assert(standard_coords == [self.x, self.y, None, None, None])

    def test_get_coordinates_points_returns_points(self):
        points = self.ug.get_coordinates_points()
        num_points = len([p for p in points])
        assert(num_points == 15)

    def test_GIVEN_missing_coord_values_WHEN_coords_THEN_missing_values_removed(self):
        x_points = np.arange(-10, 11, 5)
        y_points = np.arange(-5, 6, 5)
        y, x = np.meshgrid(y_points, x_points)
        y = np.ma.masked_array(y, np.zeros(y.shape, dtype=bool))
        y.mask[1, 2] = True

        x = Coord(x, Metadata(standard_name='latitude', units='degrees'))
        y = Coord(y, Metadata(standard_name='longitude', units='degrees'))

        coords = CoordList([x, y])

        ug = UngriddedCoordinates(coords)
        coords = ug.coords()
        for coord in coords:
            assert_that(len(coord.data), is_(14))

    def test_GIVEN_missing_coord_values_WHEN_coords_flattened_THEN_missing_values_removed(self):
        x_points = np.arange(-10, 11, 5)
        y_points = np.arange(-5, 6, 5)
        y, x = np.meshgrid(y_points, x_points)
        y = np.ma.masked_array(y, np.zeros(y.shape, dtype=bool))
        y.mask[1, 2] = True

        x = Coord(x, Metadata(standard_name='latitude', units='degrees'))
        y = Coord(y, Metadata(standard_name='longitude', units='degrees'))

        coords = CoordList([x, y])

        ug = UngriddedCoordinates(coords)
        coords = ug.coords_flattened
        for coord in coords:
            if coord is not None:
                assert_that(len(coord), is_(14))

    @skip_pandas
    def test_GIVEN_ungridded_coords_WHEN_call_as_data_frame_THEN_returns_valid_data_frame(self):

        df = self.ug.as_data_frame()

        assert_that(df['latitude'][13] == 10)
        assert_that(df['longitude'][0] == -5)


    @skip_pandas
    def test_GIVEN_ungridded_coords_with_time_WHEN_call_as_data_frame_THEN_returns_valid_data_frame(self):
        from datetime import datetime
        ug = make_regular_2d_with_time_ungridded_data()
        ug_coords = UngriddedCoordinates(ug._coords)

        df = ug_coords.as_data_frame()

        assert_that(df['latitude'][13] == 10)
        assert_that(df['longitude'][0] == -5)
        assert_that(df['longitude'][datetime(1984,8,28)] == 0)


class TestUngriddedDataList(TestCase):

    def setUp(self):
        x_points = np.arange(-10, 11, 5)
        y_points = np.arange(-5, 6, 5)
        y, x = np.meshgrid(y_points, x_points)
        x = Coord(x, Metadata(name='lat', standard_name='latitude', units='degrees'))
        y = Coord(y, Metadata(name='lon', standard_name='longitude', units='degrees'))
        data = np.reshape(np.arange(15) + 1.0, (5, 3))
        self.coords = CoordList([x, y])

        ug1 = UngriddedData(data, Metadata(standard_name='rainfall_flux', long_name="TOTAL RAINFALL RATE: LS+CONV KG/M2/S",
                                           units="kg m-2 s-1", missing_value=-999), self.coords)
        ug2 = UngriddedData(data * 0.1, Metadata(standard_name='snowfall_flux', long_name="TOTAL SNOWFALL RATE: LS+CONV KG/M2/S",
                                                 units="kg m-2 s-1", missing_value=-999), self.coords)
        self.ungridded_data_list = UngriddedDataList([ug1, ug2])

    def test_slicing(self):
        single_item = self.ungridded_data_list[1]
        assert_that(isinstance(single_item, UngriddedData))
        many_items = self.ungridded_data_list[0:1]
        assert_that(isinstance(many_items, UngriddedDataList))
        many_items = self.ungridded_data_list[0:]
        assert_that(isinstance(many_items, UngriddedDataList))

    def test_combining(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data
        another_list = UngriddedDataList([make_regular_2d_ungridded_data(), make_regular_2d_ungridded_data()])
        # Test adding
        assert_that(isinstance(self.ungridded_data_list + another_list, UngriddedDataList))
        # Test extending
        another_list.extend(self.ungridded_data_list)
        assert_that(isinstance(another_list, UngriddedDataList))
        assert_that(len(another_list) == 4)
        # Test can't add single items
        with assert_raises(TypeError):
            self.ungridded_data_list + another_list[0]

    def test_can_get_string_of_list(self):
        s = str(self.ungridded_data_list)
        assert_that(s == "UngriddedDataList: \n0: Ungridded data: rainfall_flux / (kg m-2 s-1) \n"
                         "1: Ungridded data: snowfall_flux / (kg m-2 s-1) \nCoordinates: \n  latitude\n  longitude\n")

    def test_GIVEN_data_containing_multiple_matching_coordinates_WHEN_coords_THEN_only_unique_coords_returned(self):
        unique_coords = self.ungridded_data_list.coords()
        assert_that(len(unique_coords), is_(2))
        assert_that(isinstance(unique_coords, CoordList))
        coord_names = [coord.standard_name for coord in unique_coords]
        assert_that(coord_names, contains_inanyorder('latitude', 'longitude'))

    def test_can_create_list_from_generators_and_other_iterators(self):
        from cis.test.util.mock import make_regular_2d_ungridded_data
        import itertools
        another_list = UngriddedDataList((make_regular_2d_ungridded_data(), make_regular_2d_ungridded_data()))
        assert_that(len(another_list) == 2)

        dict = {1: [make_regular_2d_ungridded_data()], 2: [make_regular_2d_ungridded_data()]}
        another_list = UngriddedDataList(itertools.chain.from_iterable(d for d in dict.values()))
        assert_that(len(another_list) == 2)

    @skip_pandas
    def test_GIVEN_multiple_ungridded_data_WHEN_call_as_data_frame_THEN_returns_valid_data_frame(self):

        df = self.ungridded_data_list.as_data_frame()

        assert_that(df['rainfall_flux'][5] == 6)
        assert_almost_equal(df['snowfall_flux'][5], 0.6)
        assert_that(df['latitude'][13] == 10)
        assert_that(df['longitude'][0] == -5)

    @skip_pandas
    def test_GIVEN_multiple_ungridded_data_with_missing_data_WHEN_call_as_data_frame_THEN_returns_valid_data_frame(self):
        d = np.reshape(np.arange(15) + 10.0, (5, 3))

        data = np.ma.masked_array(d, np.zeros(d.shape, dtype=bool))
        data.mask[1,2] = True

        ug3 = UngriddedData(data, Metadata(name='hail', long_name="TOTAL HAIL RATE: LS+CONV KG/M2/S",
                                           units="kg m-2 s-1", missing_value=-999), self.coords)

        self.ungridded_data_list.append(ug3)

        df = self.ungridded_data_list.as_data_frame()

        assert_that(df['rainfall_flux'][5] == 6)
        assert_almost_equal(df['snowfall_flux'][5], 0.6)
        assert_that(df['latitude'][13] == 10)
        assert_that(df['longitude'][0] == -5)
        assert_almost_equal(df['TOTAL HAIL RATE: LS+CONV KG/M2/S'][1], 11.0)
        assert_that(np.isnan(df['TOTAL HAIL RATE: LS+CONV KG/M2/S'][np.ravel_multi_index([1, 2], (5, 3))]))

        self.ungridded_data_list.pop()


@skip_pandas
class TestPandasHelperFunctions(TestCase):

    @raises(ValueError)
    def test_GIVEN_masked_array_WHEN_call_to_flat_ndarray_NO_copy_THEN_raises_ValueError(self):
        from cis.data_io.ungridded_data import _to_flat_ndarray
        d = np.reshape(np.arange(15) + 10.0, (5, 3))

        data = np.ma.masked_array(d, np.zeros(d.shape, dtype=bool))
        data.mask[2] = True

        _to_flat_ndarray(data, copy=False)

    def test_GIVEN_masked_array_WHEN_call_to_flat_ndarray_with_copy_THEN_returns_a_flat_copy(self):
        from cis.data_io.ungridded_data import _to_flat_ndarray
        d = np.reshape(np.arange(15) + 10.0, (5, 3))

        data = np.ma.masked_array(d, np.zeros(d.shape, dtype=bool))
        data.mask[2,2] = True

        new = _to_flat_ndarray(data, copy=True)

        assert_that(np.isnan(new[np.ravel_multi_index([2, 2], (5, 3))]))

if __name__ == '__main__':
    import nose
    nose.runmodule()



import numpy
from cis.data_io.hyperpoint import HyperPoint
from cis.data_io.hyperpoint_view import UngriddedHyperPointView
from cis.data_io.ungridded_data import LazyData
from cis.utils import fix_longitude_range


class Coord(LazyData):
    @classmethod
    def from_many_coordinates(cls, coords):
        """
        Create a single coordinate object from the concatenation of all of the coordinate objects in the input list,
        updating the shape as appropriate

        :param coords: A list of coordinate objects to be combined
        :return: A single :class:`Coord` object
        """
        from cis.utils import concatenate
        data = concatenate([ug.data for ug in coords])
        metadata = coords[0].metadata  # Use the first file as a master for the metadata...
        metadata.shape = data.shape  # But update the shape
        return cls(data, metadata, coords[0].axis)

    def __init__(self, data, metadata, axis='', data_retrieval_callback=None):
        """

        :param data:
        :param metadata:
        :param axis: A string label for the axis, e.g. 'X', 'Y', 'Z', or 'T'
        :return:
        """
        super(Coord, self).__init__(data, metadata, data_retrieval_callback)
        self.axis = axis.upper()
        # Fix an issue where cf_units cannot parse units 'deg' (should be degrees).
        if isinstance(self.units, str) and self.units == 'deg':
            self.units = 'degrees'

    def __getitem__(self, keys):
        """
        Return a COPY of the Coord with the given slice. We copy to emulate the Iris Cube behaviour
        """
        from copy import deepcopy
        # The data is just a new LazyData objects with the sliced data. Note this is a slice of the whole (concatenated)
        #  data, and will lead to post-processing before slicing.
        # TODO: We could be cleverer and figure out the right slice across the various data managers to only read the
        #  right data from disk.
        return Coord(self.data[keys].copy(), metadata=deepcopy(self.metadata), axis=self.axis)

    @property
    def points(self):
        """Alias for :func:`self.data`, to match :func:`iris.coords.Coord.points` interface

        :return: Coordinate data values
        """
        return self.data

    def __eq__(self, other):
        return other.metadata.standard_name == self.metadata.standard_name and self.metadata.standard_name != ''

    def convert_julian_to_std_time(self):
        from cis.time_util import convert_julian_date_to_std_time, cis_standard_time_unit
        # if not self.units.startswith("Julian Date"):
        #     raise ValueError("Time units must be Julian Date for conversion to an Object")
        self._data = convert_julian_date_to_std_time(self.data)
        self.units = cis_standard_time_unit

    def convert_TAI_time_to_std_time(self, ref):
        from cis.time_util import convert_sec_since_to_std_time, cis_standard_time_unit
        self._data = convert_sec_since_to_std_time(self.data, ref)
        self.units = cis_standard_time_unit

    def convert_to_std_time(self, time_stamp_info=None):
        """
        Convert this coordinate to standard time. It will use either: the units of the coordinate if it is a cf_units
        Unit, or the first word of the units, combined with the time stamp (if the timestamp is not given an error is
        thrown).

        :param time_stamp_info: the time stamp info from the file, None if it does not exist
        """
        from cis.time_util import convert_time_since_to_std_time, cis_standard_time_unit, \
            convert_time_using_time_stamp_info_to_std_time, convert_datetime_to_std_time
        from cf_units import Unit

        if isinstance(self.units, Unit):
            self._data = convert_time_since_to_std_time(self.data, self.units)
        elif str(self.units).lower().startswith('datetime'):
            self._data = convert_datetime_to_std_time(self.data)
        else:
            if time_stamp_info is None:
                raise ValueError("File must have time stamp info if converting without 'since' in units definition")
            self._data = convert_time_using_time_stamp_info_to_std_time(self.data, self.units, time_stamp_info)

        self.units = cis_standard_time_unit

    def convert_datetime_to_standard_time(self):
        from cis.time_util import convert_datetime_to_std_time, cis_standard_time_unit
        self._data = convert_datetime_to_std_time(self.data)
        self.units = cis_standard_time_unit

    def convert_standard_time_to_datetime(self):
        from cis.time_util import convert_std_time_to_datetime, cis_standard_time_unit
        if self.units == cis_standard_time_unit:
            self.data = convert_std_time_to_datetime(self.data)
            self.units = "DateTime Object"

    def set_longitude_range(self, range_start):
        """
        Confine the coordinate longitude range to 360 degrees from the :attr:`range_start` value.

        :param float range_start: Start of the longitude range
        """
        self._data = fix_longitude_range(self._data, range_start)
        self._data_flattened = None

    def copy(self, data=None):
        """
        Create a copy of this Coord object with new data so that that they can be modified without held references
        being affected. This will call any lazy loading methods in the coordinate data

        :return: Copied :class:`Coord`
        """
        from copy import deepcopy
        data = data if data is not None else numpy.ma.copy(self.data)  # Will call lazy load method
        return Coord(data, deepcopy(self.metadata), axis=deepcopy(self.axis))


class CoordList(list):
    """All the functionality of a standard :class:`list` with added :class:`Coord` context."""

    def __init__(self, *args):
        """ Given many :class:`Coord`s, return a :class:`CoordList` instance. """
        list.__init__(self, *args)

        # Check that all items in the incoming list are coords. Note that this checking
        # does not guarantee that a CoordList instance *always* has just coords in its list as
        # the append & __getitem__ methods have not been overridden.
        if not all([isinstance(coord, Coord) for coord in self]):
            raise ValueError('All items in list_of_coords must be Coord instances.')

    def append(self, other):
        """
        Safely add a new coordinate object to the list, this checks for a unique :attr:`axis` and :attr:`standard_name`.

        :param :class:`Coord` other: Other coord to add
        :raises DuplicateCoordinateError: If the coordinate is not unique in the list
        """
        from cis.exceptions import DuplicateCoordinateError
        if any([other == item for item in self]):
            raise DuplicateCoordinateError()
        super(CoordList, self).append(other)

    def get_coords(self, name_or_coord=None, standard_name=None, long_name=None, attributes=None, axis=None,
                   var_name=None):
        """
        Return a list of coordinates in this :class:`CoordList` fitting the given criteria. This is deliberately very
        similar to :func:`Cube.coords()` to maintain a similar interface and because the functionality is similar. There
        is no distinction between dimension coordinates and auxiliary coordinates here though.

        :param name_or_coord: This should be either: The standard name or long name or default name of the desired
         coordinate; Or, a :class:`Coord` instance whose metadata should be used for the search criteria (note that
         currently only the standard name is compared). If None, does not check for name. Also see, :attr:`Cube.name`.
        :type name: string or None or :class:`Coord`
        :param standard_name: The CF standard name of the desired coordinate. If None, does not check for standard name.
        :type standard_name: string or None
        :param long_name: An unconstrained description of the coordinate. If None, does not check for long_name.
        :type long_name: string or None
        :param attributes: A dictionary of attributes desired on the coordinates. If None, does not check for attributes
        :type attributes: dict or None
        :param axis: The desired coordinate axis, see :func:`iris.util.guess_coord_axis`. If None, does not check for
         axis. Accepts the values 'X', 'Y', 'Z' and 'T' (case-insensitive).
        :type axis: string or None
        :param var_name: The name of the variable which the coordinate was read from. If None, does not check for long_name.
        :type var_name: string or None

        :return: A :class:`CoordList` of coordinates fitting the given criteria
        """
        import six
        from collections import Mapping
        coords = self

        if isinstance(name_or_coord, six.string_types):
            name = name_or_coord
            coord = None
        else:
            name = None
            coord = name_or_coord

        if name is not None:
            coords = [coord_ for coord_ in coords if coord_.name() == name]

        if standard_name is not None:
            coords = [coord_ for coord_ in coords if coord_.standard_name == standard_name]

        if long_name is not None:
            coords = [coord_ for coord_ in coords if coord_.long_name == long_name]

        if axis is not None:
            axis = axis.upper()
            coords = [coord_ for coord_ in coords if coord_.axis == axis]

        if var_name is not None:
            coords = [coord_ for coord_ in coords if coord_.var_name == var_name]

        if attributes is not None:
            if not isinstance(attributes, Mapping):
                raise ValueError(
                    'The attributes keyword was expecting a dictionary type, but got a %s instead.' % type(attributes))
            coords = [coord_ for coord_ in coords if all(
                            k in coord_.attributes and coord_.attributes[k] == v for k, v in attributes.items())]

        if coord is not None:
            coords = [coord_ for coord_ in coords if coord_ == coord]

        return coords

    def get_coord(self, name_or_coord=None, standard_name=None, long_name=None, attributes=None, axis=None,
                  var_name=None):
        """
        Return a single coord fitting the given criteria. This is deliberately very
        similar to :func:`Cube.coord()` method to maintain a similar interface and because the functionality is similar.
        There is no distinction between dimension coordinates and auxilliary coordinates here though.

        :param name_or_coord: This should be either: The standard name or long name or default name of the desired
         coordinate; Or, a :class:`Coord` instance whose metadata should be used for the search criteria (note that
         currently only the standard name is compared). If None, does not check for name. Also see, :attr:`Cube.name`.
        :type name: string or None or :class:`Coord`
        :param standard_name: The CF standard name of the desired coordinate. If None, does not check for standard name.
        :type standard_name: string or None
        :param long_name: An unconstrained description of the coordinate. If None, does not check for long_name.
        :type long_name: string or None
        :param attributes: A dictionary of attributes desired on the coordinates. If None, does not check for attributes
        :type attributes: dict or None
        :param axis: The desired coordinate axis, see :func:`iris.util.guess_coord_axis`. If None, does not check for
         axis. Accepts the values 'X', 'Y', 'Z' and 'T' (case-insensitive).
        :type axis: string or None
        :param var_name: The name of the variable which the coordinate was read from. If None, does not check for long_name.
        :type var_name: string or None

        :raises CoordinateNotFoundError: If the arguments given do not result in precisely
         1 coordinate being matched.
        :return: A single :class:`Coord`.

        """
        from cis.exceptions import CoordinateNotFoundError
        coords = self.get_coords(name_or_coord=name_or_coord, standard_name=standard_name, long_name=long_name,
                                 attributes=attributes, axis=axis, var_name=var_name)
        if len(coords) == 0:  # If we found none by name, try with standard name only
            coords = self.get_coords(standard_name=name_or_coord)

        if len(coords) > 1:
            msg = 'Expected to find exactly 1 coordinate, but found %s. They were: %s.' \
                  % (len(coords), ', '.join(coord.name() for coord in coords))
            raise CoordinateNotFoundError(msg)
        elif len(coords) == 0:
            bad_name = str(name_or_coord) or standard_name or long_name or axis or ''
            msg = 'Expected to find exactly 1 %s coordinate, but found none.' % bad_name
            raise CoordinateNotFoundError(msg)

        return coords[0]

    def get_coordinates_points(self):
        all_coords = self.find_standard_coords()
        flattened_coords = [(c.data_flattened if c is not None else None) for c in all_coords]
        return UngriddedHyperPointView(flattened_coords, None)

    def find_standard_coords(self):
        """Constructs a list of the standard coordinates.
        The standard coordinates are latitude, longitude, altitude, air_pressure and time; they occur in the return
        list in this order.

        :return: :class:`list` of coordinates or None if coordinate not present
        """
        from cis.exceptions import CoordinateNotFoundError

        ret_list = []

        for name in HyperPoint.standard_names:
            try:
                coord = self.get_coord(standard_name=name)
            except CoordinateNotFoundError:
                coord = None
            ret_list.append(coord)

        return ret_list

    def copy(self):
        """
        Create a copy of this CoordList object with new data so that that they can
        be modified without held references being affected. This will call any lazy loading methods in the coordinate
        data

        :return: Copied :class:`CoordList`
        """
        copied = CoordList()
        for coord in self:
            copied.append(coord.copy())
        return copied



"""Tests for parse_datetime module
"""
from nose.tools import istest, raises, assert_almost_equal, eq_
from cis.parse_datetime import _parse_partial_datetime, parse_as_number_or_datetime, \
                               parse_datetimestr_delta_to_float_days, parse_datetimestr_to_std_time
from cis.time_util import PartialDateTime


# Tests for parse_datetime
@istest
def parse_datetime_can_parse_year():
    dt = _parse_partial_datetime('2010')
    assert (dt == PartialDateTime(2010))


@istest
def parse_datetime_can_parse_year_month():
    dt = _parse_partial_datetime('2010-07')
    print(dt)
    assert (dt == PartialDateTime(2010, 7))


@istest
def parse_datetime_can_parse_date():
    dt = _parse_partial_datetime('2010-07-01')
    assert (dt == PartialDateTime(2010, 7, 1))


@istest
def parse_datetime_can_parse_date_hour():
    dt = _parse_partial_datetime('2010-07-01T13')
    assert (dt == PartialDateTime(2010, 7, 1, 13))


@istest
def parse_datetime_can_parse_date_hour_min():
    dt = _parse_partial_datetime('2010-07-01T13:27')
    assert (dt == PartialDateTime(2010, 7, 1, 13, 27))


@istest
def parse_datetime_can_parse_date_hour_min_sec():
    dt = _parse_partial_datetime('2010-07-01T13:27:43')
    assert (dt == PartialDateTime(2010, 7, 1, 13, 27, 43))


@istest
def parse_datetime_can_parse_date_hour_min_sec_no_leading_zeros():
    dt = _parse_partial_datetime('2010-3-4T5:6:7')
    assert (dt == PartialDateTime(2010, 3, 4, 5, 6, 7))


@istest
def parse_datetime_can_parse_date_time_with_space_separator():
    dt = _parse_partial_datetime('2010-07-01 13:27:43')
    assert (dt == PartialDateTime(2010, 7, 1, 13, 27, 43))


@istest
def parse_datetime_can_parse_date_time_with_colon_separator():
    dt = _parse_partial_datetime('2010-07-01:13:27:43')
    assert (dt == PartialDateTime(2010, 7, 1, 13, 27, 43))


# parse_datetime: Parse errors
@istest
@raises(ValueError)
def parse_datetime_raises_error_if_invalid_character_in_year():
    dt = _parse_partial_datetime('2X10')


@istest
@raises(ValueError)
def parse_datetime_raises_error_if_time_but_incomplete_date():
    dt = _parse_partial_datetime('2010-10T12:00')


@istest
@raises(ValueError)
def parse_datetime_raises_error_if_too_many_date_components():
    dt = _parse_partial_datetime('2010-10-05-06')


@istest
@raises(ValueError)
def parse_datetime_raises_error_if_too_many_time_components():
    dt = _parse_partial_datetime('2010-10-05T12:01:02:03')


# parse_datetime: Strings that parse correctly but correspond to invalid date/times
@istest
@raises(ValueError)
def parse_datetime_raises_error_if_invalid_month():
    dt = _parse_partial_datetime('2010-13')


@istest
@raises(ValueError)
def parse_datetime_raises_error_if_invalid_day():
    dt = _parse_partial_datetime('2010-06-31')


# Tests for parse_as_number_or_datetime
@istest
def parse_as_number_or_datetime_can_parse_date_as_datetime():
    from datetime import datetime
    dt = parse_as_number_or_datetime('2010-07-01')
    assert (dt == datetime(2010, 7, 1))


@istest
def parse_as_number_or_datetime_can_parse_integer():
    dt = parse_as_number_or_datetime('2010')
    assert (dt == 2010)


@istest
def parse_as_number_or_datetime_can_parse_float():
    dt = parse_as_number_or_datetime('12.345')
    assert (dt == 12.345)


@istest
def test_that_can_parse_time_deltas():
    delta = parse_datetimestr_delta_to_float_days("P2y15m3dT5M10H3S")
    assert_almost_equal(1183.420173611111, delta)


@istest
@raises(ValueError)
def test_that_raise_an_error_when_datetimestr_delta_is_invalid():
    parse_datetimestr_delta_to_float_days("some wierd string")

@istest
def test_that_can_parse_datetimestr_to_obj():
    from cis.time_util import convert_datetime_to_std_time
    import datetime as dt
    # when not specifying the hours, minutes or seconds, 0 is used
    eq_(parse_datetimestr_to_std_time("2010-02-05 02:15:45"),
        convert_datetime_to_std_time(dt.datetime(2010, 2, 5, 2, 15, 45)))
    eq_(parse_datetimestr_to_std_time("2010-02-05 02:15"),
        convert_datetime_to_std_time(dt.datetime(2010, 2, 5, 2, 15, 0)))
    eq_(parse_datetimestr_to_std_time("2010-02-05 02"),
        convert_datetime_to_std_time(dt.datetime(2010, 2, 5, 2, 0, 0)))
    eq_(parse_datetimestr_to_std_time("2010-02-05"),
        convert_datetime_to_std_time(dt.datetime(2010, 2, 5, 0, 0, 0)))

    # GOTCHA: when not specifying an element of a date (i.e. the year, month or day), the current date is used
    now = dt.datetime.now()
    eq_(parse_datetimestr_to_std_time("2010-02-05"),
        convert_datetime_to_std_time(dt.datetime(2010, 2, 5)))
    eq_(parse_datetimestr_to_std_time("2010-12"),
        convert_datetime_to_std_time(dt.datetime(2010, 12, now.day)))
    eq_(parse_datetimestr_to_std_time("2010-"),
        convert_datetime_to_std_time(dt.datetime(2010, now.month, now.day)))


if __name__ == '__main__':
    import nose

    nose.runmodule()



"""
Plotting integration tests. These tests use matplotlib to actually check the image output is 'similar' to some reference
image. The tolerance is defined in the test_plotting script and based onthe default used in iris 1.8.0.
"""


from matplotlib.ticker import LogFormatter, is_close_to_int, nearest_long
from matplotlib import rcParams
import math


class LogFormatterMathtextSpecial(LogFormatter):
    """
    Special formatter for color log axis, using notation such as 2.3 x 10 ^ 4.
    This is a modified version of LogFormatterMathtext supplied by matplotlib.
    """

    def __call__(self, x, pos=None):
        """Return the format for tick val *x* at position *pos*"""
        b = self._base
        usetex = rcParams['text.usetex']

        # only label the decades
        if x == 0:
            if usetex:
                return '$0$'
            else:
                return '$\mathdefault{0}$'

        fx = math.log(abs(x)) / math.log(b)
        is_decade = is_close_to_int(fx)
        fx = math.floor(fx)

        decimal_part = x / (b**fx)

        sign_string = '-' if x < 0 else ''

        # use string formatting of the base if it is not an integer
        if b % 1 == 0.0:
            base = '%d' % b
        else:
            base = '%s' % b

        if not is_decade and self.labelOnlyBase:
            return ''
        elif not is_decade:
            if usetex:
                return '$%s\\times%s%s^{%d}$' % ('%1.1f' % decimal_part, sign_string, base, fx)
            else:
                return '$\mathdefault{%s\\times%s%s^{%d}}$' % ('%1.1f' % decimal_part, sign_string, base, fx)
        else:
            if usetex:
                return r'$%s%s^{%d}$' % (sign_string, base, nearest_long(fx))
            else:
                return r'$\mathdefault{%s%s^{%d}}$' % (sign_string, base, nearest_long(fx))



# The RegularGridInterpolator class was adapted from the SciPy original here:
# https://github.com/scipy/scipy/blob/master/scipy/interpolate/interpolate.py

# Copyright (c) 2001, 2002 Enthought, Inc.
# All rights reserved.
#
# Copyright (c) 2003-2013 SciPy Developers.
# All rights reserved.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
#  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


# There is no algorithmic change, just a restructuring to allow caching of the weights for calculating many
#  interpolations of different datasets using the same points, and support for hybrid coordinates.
# The two helper functions extend_circular_coord and _data are also taken from SciPy as the interpolate module is
#  deprecated since Iris 1.10.
import numpy as np


def extend_circular_coord(coord, points):
    """
    Return coordinates points with a shape extended by one
    This is common when dealing with circular coordinates.

    """
    modulus = np.array(coord.units.modulus or 0,
                       dtype=coord.dtype)
    points = np.append(points, points[0] + modulus)
    return points


def extend_circular_data(data, coord_dim):
    coord_slice_in_cube = [slice(None)] * data.ndim
    coord_slice_in_cube[coord_dim] = slice(0, 1)

    data = np.append(data,
                     data[tuple(coord_slice_in_cube)],
                     axis=coord_dim)
    return data


class GriddedUngriddedInterpolator(object):

    def __init__(self, _data, sample, method='lin', missing_data_for_missing_sample=False):
        """
        Prepare an interpolation over the grid defined by a GriddedData source onto an UngriddedData sample.

        Indices are calculated but no interpolation is performed until the resulting object is called. Note that
        if the source contains a hybrid vertical coordinate these ARE interpolated to find a single vertical index.

        :param GriddedData _data: The source data, only the coordinates are used from this at initialisation.
        :param UngriddedData sample: The points to sample the source data at.
        :param str method: The interpolation method to use (either 'linear' or 'nearest'). Default is 'linear'.
        """
        from cis.utils import move_item_to_end
        coords = []
        grid_points = []
        self._circular_coord_dims = []
        self._decreasing_coord_dims = []

        # If we have hybrid coordinates then ensure that the vertical dimension of the cube is last, as we will be
        #  performing partial interpolations to pull out vertical columns - which only works if the vertical dim is last
        # We do this first so that the coordinates are pulled out in the correct, new, order.
        if len(_data.aux_factories) > 0:
            vertical_dim = _data.coord_dims(_data.aux_factories[0].sigma)[0]
            self._data_transpose = move_item_to_end(range(_data.ndim), vertical_dim)
        else:
            self._data_transpose = list(range(_data.ndim))

        # Unfortunately transpose works in place, so make a copy of the data first
        data = _data.copy()
        data.transpose(self._data_transpose)

        # Remove any tuples in the list that do not correspond to a dimension coordinate in the cube 'data'.
        for coord in data.coords(dim_coords=True):
            if len(sample.coords(standard_name=coord.name())) > 0:
                coords.append(coord.name())
                # The circular coord must be a dim coord so we pop the value out of the tuple
                coord_dim = data.coord_dims(coord)[0]
                coord_points = coord.points

                decreasing = (coord_points.size > 1 and
                              coord_points[1] < coord_points[0])
                if decreasing:
                    self._decreasing_coord_dims.append(decreasing)
                    coord_points = coord_points[::-1]

                if getattr(coord, 'circular', False):
                    self._circular_coord_dims.append(coord_dim)
                    coord_points = extend_circular_coord(coord, coord_points)

                grid_points.append(coord_points)

        sample_points = [sample.coord(c).data_flattened for c in coords]

        if len(data.coords('altitude', dim_coords=False)) > 0 and sample.coords(standard_name='altitude'):
            hybrid_dims = data.coord_dims(data.coord('altitude'))
            hybrid_coord = data.coord('altitude').points
            for coord_dim in self._circular_coord_dims:
                # Lookup the dimension on the hybrid data array
                hybrid_coord_dim = hybrid_dims.index(coord_dim)
                hybrid_coord = extend_circular_data(hybrid_coord, hybrid_coord_dim)
            sample_points.append(sample.coord(standard_name="altitude").data_flattened)
        elif len(data.coords('air_pressure', dim_coords=False)) > 0 and sample.coords(standard_name='air_pressure'):
            hybrid_coord = data.coord('air_pressure').points
            for coord_dim in self._circular_coord_dims:
                hybrid_coord = extend_circular_data(hybrid_coord, coord_dim)
            hybrid_dims = data.coord_dims(data.coord('air_pressure'))
            sample_points.append(sample.coord(standard_name="air_pressure").data_flattened)
        else:
            hybrid_coord = None
            hybrid_dims = None

        if len(sample_points) != len(data.shape):
            raise ValueError("Sample points do not uniquely define gridded data source points, invalid "
                             "dimenions: {} and {} respectively".format(len(sample_points), len(data.shape)))

        # Check if we want to sample missing points
        if missing_data_for_missing_sample and hasattr(sample.data, 'mask'):
            self.missing_mask = sample.data.mask.flatten()
            sample_points = [p[~self.missing_mask] for p in sample_points]
        else:
            self.missing_mask = None

        self._interp = _RegularGridInterpolator(grid_points, sample_points,
                                                hybrid_coord=hybrid_coord, hybrid_dims=hybrid_dims, method=method)

    def _get_dims_order(self, data, coords):
        """
        Return the dims with the vertical coord last. There must be a nicer way of doing this...
        """
        # Find the only dimension coordinate we haven't yet accounted for - this must be the vertical
        vertical_coord = list(set([c.name() for c in data.coords(dim_coords=True)]).difference(set(coords)))[0]
        vertical_dim = data.coord_dims(vertical_coord)[0]
        # The dim order is just the range of the dims with the vertical dim moved to the end.
        self._data_transpose.pop(vertical_dim)
        self._data_transpose.append(vertical_dim)

    def _account_for_inverted(self, data):
        dim_slices = [slice(None)] * data.ndim
        for dim in self._decreasing_coord_dims:
                dim_slices[dim] = slice(-1, None, -1)
        data = data[dim_slices]
        return data

    def __call__(self, data, fill_value=np.nan, extrapolate=False):
        """
         Perform the prepared interpolation over the given data GriddedData object - this assumes that the coordinates
          used to initialise the interpolator are identical as those in this data object.

        If extrapolate is True then fill_value is ignored (since there will be no invalid values).

        :param GriddedData data: Data values to interpolate
        :param float fill_value: The fill value to use for sample points outside of the bounds of the data
        :param bool extrapolate: Extrapolate points outside the bounds of the data? Default False.
        :return ndarray: Interpolated values.
        """
        if extrapolate:
            fill_value = None

        # Apply a transpose if we need to so that the indices line-up correctly
        data_array = data.data.transpose(self._data_transpose)
        # Account for any circular coords present
        for dim in self._circular_coord_dims:
            data_array = extend_circular_data(data_array, dim)

        data_array = self._account_for_inverted(data_array)

        result = self._interp(data_array, fill_value=fill_value)

        if self.missing_mask is not None:
            # Pack the interpolated values back into the original shape
            expanded_result = np.ma.masked_array(np.zeros(self.missing_mask.shape), mask=self.missing_mask.copy(),
                                                 fill_value=fill_value)
            expanded_result[~self.missing_mask] = result
            result = expanded_result

        return result


def _ndim_coords_from_arrays(points, ndim=None):
    """
    Convert a tuple of coordinate arrays to a (..., ndim)-shaped array.

    """
    if isinstance(points, tuple) and len(points) == 1:
        # handle argument tuple
        points = points[0]
    if isinstance(points, tuple):
        p = np.broadcast_arrays(*points)
        for j in range(1, len(p)):
            if p[j].shape != p[0].shape:
                raise ValueError(
                    "coordinate arrays do not have the same shape")
        points = np.empty(p[0].shape + (len(points),), dtype=float)
        for j, item in enumerate(p):
            points[..., j] = item
    else:
        points = np.asanyarray(points)
        # XXX Feed back to scipy.
        if points.ndim <= 1:
            if ndim is None:
                points = points.reshape(-1, 1)
            else:
                points = points.reshape(-1, ndim)
    return points


class _RegularGridInterpolator(object):
    """
    Interpolation on a regular grid in arbitrary dimensions
    The data must be defined on a regular grid; the grid spacing however may be
    uneven.  Linear and nearest-neighbour interpolation are supported. After
    setting up the interpolator object, the interpolation can be performed for
    multiple data arrays with cached indices - this assumes they have the same
    coordinates.
    """
    # this class is based on code originally programmed by Johannes Buchner,
    # see https://github.com/JohannesBuchner/regulargrid

    def __init__(self, coords, points, hybrid_coord=None, hybrid_dims=None, method="lin"):
        """
        Initialise the itnerpolator - this will calculate and cache the indices of the interpolation. It will
        also interpolate the hybrid coordinate if needed to determine a unique vertical index.

        :param iterable coords: The coords defining the regular grid in n dimensions. Should be a tuple of ndarrays
        :param ndarray points: The points to sample the gridded data at.
        :param ndarray hybrid_coord: An (optional) array describing a single vertical hybrid coordinate
        :param iterable hybrid_dims: The grid dimensions over which the hybrid coordinate is defined
        :param str method: The method of interpolation to perform. Supported are "linear" and "nearest". Default is
        "linear".
        """
        if method == "lin":
            self._interp = self._evaluate_linear
        elif method == "nn":
            self._interp = self._evaluate_nearest
        else:
            raise ValueError("Method '%s' is not defined" % method)

        for i, c in enumerate(coords):
            if not np.all(np.diff(c) > 0.):
                raise ValueError("The points in dimension %d must be strictly "
                                 "ascending" % i)
            if not np.asarray(c).ndim == 1:
                raise ValueError("The points in dimension %d must be "
                                 "1-dimensional" % i)
        self.grid = tuple([np.asarray(c) for c in coords])

        if hybrid_coord is None:
            ndim = len(self.grid)
            points = np.vstack(points).T
            if points.shape[-1] != len(self.grid):
                raise ValueError("The requested sample points have dimension "
                                 "%d, but the interpolation grid has "
                                 "dimension %d" % (points.shape[1], ndim))
        else:
            ndim = len(self.grid)
            points = _ndim_coords_from_arrays(points, ndim=ndim)

        if hybrid_coord is not None:

            # Firstly interpolate over all of the dimensions except the vertical (which will always be the last...)
            self.indices, self.norm_distances, self.out_of_bounds = \
                self._find_indices(points[:-1], self.grid)

            # Find the dims to interpolate over for the hybrid coord
            hybrid_interp_dims = hybrid_dims[:-1]
            hybrid_indices = [self.indices[i] for i in hybrid_interp_dims]

            # Find all of the interpolated vertical columns (one for each point)
            v_coords = self._interp(hybrid_coord, hybrid_indices, self.norm_distances)

            self.indices.append(np.zeros(len(points.T), dtype=self.indices[0].dtype))
            self.norm_distances.append(np.zeros(len(points.T)))

            # Calculate and store the vertical index and weight for each point based on the interpolated vertical
            # column
            # TODO: This loop could be a bottle-neck for many points, maybe move to Cython?
            for i, p in enumerate(points.T):
                vert_index, vert_norm_distance, vert_out_of_bounds = self._find_vertical_index(p[-1], v_coords[i])
                self.indices[-1][i] = vert_index
                self.norm_distances[-1][i] = vert_norm_distance
                self.out_of_bounds[i] += vert_out_of_bounds

        else:
            self.indices, self.norm_distances, self.out_of_bounds = self._find_indices(points.T, self.grid)

    def __call__(self, values, fill_value=np.nan):
        """
        Interpolation of values at cached coordinates

        :param ndarray values: The data on the regular grid in n dimensions
        :param float fill_value: If provided, the value to use for points outside of the interpolation domain. If None,
        values outside the domain are extrapolated.
        :return ndarray: The interpolated values
        """
        if not hasattr(values, 'ndim'):
            # allow reasonable duck-typed values
            values = np.asarray(values)

        if len(self.grid) > values.ndim:
            raise ValueError("There are %d point arrays, but values has %d "
                             "dimensions" % (len(self.grid), values.ndim))

        if hasattr(values, 'dtype') and hasattr(values, 'astype'):
            if not np.issubdtype(values.dtype, np.inexact):
                values = values.astype(float)

        if fill_value is not None:
            fill_value_dtype = np.asarray(fill_value).dtype
            if (hasattr(values, 'dtype') and not
                    np.can_cast(fill_value_dtype, values.dtype,
                                        casting='same_kind')):
                raise ValueError("fill_value must be either 'None' or "
                                 "of a type compatible with values")

        for i, p in enumerate(self.grid):
            if not values.shape[i] == len(p):
                raise ValueError("There are %d points and %d values in "
                                 "dimension %d" % (len(p), values.shape[i], i))

        result = self._interp(values, self.indices, self.norm_distances)

        if fill_value is not None:
            result = np.ma.array(result, mask=self.out_of_bounds, fill_value=fill_value)

        return result

    @staticmethod
    def _evaluate_linear(values, indices, norm_distances):
        from itertools import product
        # slice for broadcasting over trailing dimensions in self.values
        vslice = (slice(None),) + (None,)*(values.ndim - len(indices))

        # find relevant values
        # each i and i+1 represents a edge
        edges = product(*[[i, i + 1] for i in indices])
        value = 0.
        for edge_indices in edges:
            weight = 1.
            for ei, i, yi in zip(edge_indices, indices, norm_distances):
                weight *= np.where(ei == i, 1 - yi, yi)
            value += np.ma.asarray(values[edge_indices]) * weight[vslice]
        return value

    @staticmethod
    def _evaluate_nearest(values, indices, norm_distances):
        idx_res = []
        for i, yi in zip(indices, norm_distances):
            idx_res.append(np.where(yi <= .5, i, i + 1))
        return values[idx_res]

    @staticmethod
    def _find_indices(points, coords):
        # find relevant edges between which xi are situated
        indices = []
        # compute distance to lower edge in unity units
        norm_distances = []
        # check for out of bounds xi
        out_of_bounds = np.zeros((points.shape[1]), dtype=bool) if isinstance(points, np.ndarray) else 0
        # iterate through dimensions
        for x, coord in zip(points, coords):
            i = np.searchsorted(coord, x) - 1
            i[i < 0] = 0
            i[i > coord.size - 2] = coord.size - 2

            norm_distances.append((x - coord[i]) / (coord[i + 1] - coord[i]))
            indices.append(i)

            out_of_bounds += x < coord[0]
            out_of_bounds += x > coord[-1]
        return indices, norm_distances, out_of_bounds

    @staticmethod
    def _find_vertical_index(point, coord):
        i = np.searchsorted(coord, point) - 1
        if i < 0:
            i = 0
        if i > coord.size - 2:
            i = coord.size - 2
        norm_distances = ((point - coord[i]) / (coord[i + 1] - coord[i]))

        out_of_bounds = point < coord[0] or point > coord[-1]
        return i, norm_distances, out_of_bounds



import unittest

from cis.test.integration_test_data import valid_aeronet_filename, valid_aeronet_variable
from cis.data_io.aeronet import load_aeronet
from nose.tools import assert_almost_equal, raises, assert_equal
from cis.exceptions import InvalidVariableError


class TestAeronet(unittest.TestCase):

    def test_aeronet_variable_name_parsing(self):
        from cf_units import Unit
        from cis import read_data_list

        aeronet_data = read_data_list(valid_aeronet_filename, ["AOT_440", "Water(cm)"])

        assert_equal(aeronet_data[0].name(), "AOT_440")
        assert_equal(aeronet_data[0].units, Unit('1'))
        assert_equal(aeronet_data[1].name(), "Water(cm)")
        assert_equal(aeronet_data[1].var_name, "Water")
        assert_equal(aeronet_data[1].units, Unit('cm'))

    def test_aeronet_time_parsing(self):
        # 1.8s
        from datetime import datetime
        from cis.time_util import cis_standard_time_unit as ct

        aeronet_data = load_aeronet(valid_aeronet_filename, [valid_aeronet_variable])

        assert_almost_equal(aeronet_data['datetime'][0], ct.date2num(datetime(2003, 9, 25, 6, 47, 9)))
        assert_almost_equal(aeronet_data['datetime'][5], ct.date2num(datetime(2003, 9, 25, 7, 10, 37)))
        assert_almost_equal(aeronet_data['datetime'][76], ct.date2num(datetime(2003, 9, 27, 13, 28, 2)))

    @raises(InvalidVariableError)
    def test_aeronet_missing_variable(self):
        aeronet_data = load_aeronet(valid_aeronet_filename, ['invalid_variable'])

